from nuctools import *
import numpy as np
import h5py as h5
from numpy.testing import assert_array_equal,assert_allclose
import os,sys


def test_inc():
	assert inc(4) == 5

def test_get_montrig():
	sample = "test"
	i = np.zeros(8)
	j = np.zeros(2)
	with h5.File("tests/{}_master.h5".format(sample),"w") as hdf:
	    temp_mon = np.array([100.0,200.0,300.0,0.0,0.0,0.0,0.0,400.0])
	    temp_trig = np.array([[3e8,3.1e8],[0,0]])
	    temp_data = np.array([1,2,3,4,5])
	    mon = hdf.create_dataset("test/CYCLE0001/MON",data=temp_mon)
	    trig = hdf.create_dataset("test/CYCLE0001/TRIG",data=temp_trig)
	    data_g = hdf.create_dataset("test/CYCLE0001/DATA/GOOD",data=temp_data)
	    data_t = hdf.create_dataset("test/CYCLE0001/DATA/TRUNC",data=temp_data)
	    mon2 = hdf.create_dataset("test/CYCLE0002/MON",data=temp_mon)
	    trig2 = hdf.create_dataset("test/CYCLE0002/TRIG",data=temp_trig)
	    data_g2 = hdf.create_dataset("test/CYCLE0002/DATA/GOOD",data=temp_data)
	    data_t2 = hdf.create_dataset("test/CYCLE0002/DATA/TRUNC",data=temp_data)
    
	with h5.File("tests/{}_master.h5".format(sample),"r") as hdf:
	    for cycle in hdf[sample+"/"]:
	        print(cycle)
	        mon_vector = hdf[sample+"/"+cycle+"/MON"]
	        trig_vector = hdf[sample+"/"+cycle+"/TRIG"][0]
	        datag_array = hdf[sample+"/"+cycle+"/DATA/GOOD"]
	        datat_array = hdf[sample+"/"+cycle+"/DATA/TRUNC"]
	        sum_detcts = len(datag_array)+len(datat_array)
	        i+=mon_vector
	        j+=trig_vector

	manual_montrig = np.concatenate((i,j))
	
	fnc_montrig = get_montrig("tests/",sample)

	assert_array_equal(manual_montrig,fnc_montrig)

def test_count_rate():
	tof_arr = np.array([[1,2,3],[20,10,5]])
	trig = 3e8
	tau = .6
	bw = 1.0

	# convert to cps
	cps = tof_arr[1]/bw/1e-6/trig
	# error cps
	e_cps = np.sqrt((np.sqrt(tof_arr[1])/bw/trig/1e-6)**2)
	# dead-time correction
	dead_time_corr = dtc(tof_arr[1],tau,trig,bw)
	# dead-time corrected cps
	dc_cps = cps*dead_time_corr[0]
	# error in dc cps
	e_dc_cps = np.sqrt((e_cps/cps)**2+(dead_time_corr[1]/dead_time_corr[0])**2)*dc_cps
	a = np.array([tof_arr[0],dc_cps,e_dc_cps])

	fcn_cps = count_rate(tof_arr,trig,tau) 

	assert_allclose(a,fcn_cps)

def test_dtc():
	counts = np.linspace(10,1,10)
	dt = 1.125
	w = 0.5
	trigs = 100

	lc = len(counts)
	test_dtcf = np.zeros(lc)
	test_ddtcf = np.zeros(lc)
	# SUM from Dr. Danon's dead_time.pdf (maybe his thesis...)
	SUM = 0.75*counts[0:lc-2] + counts[1:lc-1] + 0.5*counts[2:lc]
	# the error on the sum
	dSUM = np.sqrt(0.75**2*counts[0:lc-2]+counts[1:lc-1]+0.5**2*counts[2:lc])

	test_dtcf[2:lc] = 1/(1-SUM/trigs)
	test_ddtcf[2:lc] = dSUM/trigs/(1-SUM/trigs)**2

	dtcf,ddtcf = dtc(counts,dt,trigs,w)

	assert_allclose(test_dtcf,dtcf)
	assert_allclose(test_ddtcf,ddtcf)




import numpy as np
import pandas as pd
from . import tof_tools as tt 
from . import math_tools as mt
from . import sam_tools as st

__all__ = ['Trans','calc_cov','write_covar_file','read_covar_file']

# TODO: create sample/open objects

class Trans:
    """
    This is a class designed to hold all the data and methods necessary to calculate
    neutron transmission from a time-of-flight measurement at RPI.
    
    Parameters
    ----------
    
    open_data_name : string
        Full path name for the file containing the OPEN data
    open_mon_name : string
        Full path name for the file containing the OPEN monitors
    sample_data_name : string
        Full path name for the file containing the SAMPLE data
    sample_mon_name : string
        Full path name for the file containing the SAMPLE data
    datafile_names : list, optional
        List of strings for the column names for the data. OPEN and
        SAMPLE DATA should typically have the same structure
    tau : float, optional
        Tau is the dead time in micro seconds
    correct_deadtime : bool, optional
        Whether the deadtime of the system needs to be corrected for 
        in the counts data
    mon_index : int, optional
        Which monitor to use. Historically it is 0-10, with the Linac 
        triggers being 0,1
    bkgfunction : string, optional
        Options are: \'power_law\', \'double_exp\', \'double_power_law\',
        and \'exp_log\'.
        There are only four background shapes, the coefficients are 
        controlled with o_bkgcoeff, and s_bkgcoeff parameters
    o_bkgcoeff : list, optional
        Coefficients for the either of the two background fucntions 
        programmed for the OPEN data. power_law needs 3 coeff's, 
        double_exp needs 5 coeffecients, double_power_law requires
        5 coefficients, exp_log requires 4 coeff's
    s_bkgcoeff : list, optional
        Coefficients for the either of the two background fucntions 
        programmed for the SAMPLE data. power_law needs 3 coeff's, 
        double_exp needs 5 coeffecients, double_power_law requires
        5 coefficients, exp_log requires 4 coeff's
    factor : 1-d numpy array, optional
        Compression factors for the tof groups, 1st compression factor
        corresponds to time of flights below c_pt[0], 2nd corresponds to tofs above
        c_pt[0] and below c_pt[1], and so on until end of range is found.
    c_pts : 1-d numpy array, optional
        Compression time of flights that seperate the bins that are to be grouped.
        List points from lowest to highest (should re-write to sort and remove user
        capability to fail.)
    FP : float, optional
        Flight path needed to convert the time of flight values of bins to
        energies for the compression point selection.
    samp_bkg_err : float, optional
        The fractional error for every point on the background. Default is
        0.2 (20%). This is a temporary way to conservatively estimate error 
        on the transmission.
    open_bkg_err : float, optional
        The fractional error for every point on the background. Default is
        0.2 (20%). This is a temporary way to conservatively estimate error 
        on the transmission.
    optstat_group : bool, optional
        Whether or not you want to use an optimized grouper 
    D : float, optional
    res_per_bin : int, optional
    E_cpts : array-like
    
    Attributes
    ----------
    
    open_data : DataFrame
        DataFrame of time-of-flight (tof), bin widths (bin_width), counts 
        (counts), count rate (cps) and error on the count rate (dcps) from
        method calc_count_rate, and background (bkg) for the OPEN. All 
        time values should be in units of microseconds.
    sample_data : DataFrame
        DataFrame of time-of-flight (tof), bin widths (bin_width), counts 
        (counts), count rate (cps) and error on the count rate (dcps) from
        method calc_count_rate, and background (bkg) for the SAMPLE. All 
        time values should be in units of microseconds.
        All time values should be in units of microseconds. Also contains
        the count rate and error in the count rate as given by the method
        calc_count_rate
    open_mon : DataFrame
        Monitor detector counts for OPEN. 1-d list of integers
    sample_mon : DataFrame
        Monitor detector counts for SAMPLE. 1-d list of integers
    open_trig : int
        The triggers used to calculate the count rate for the OPEN. 
        Typically this should be the number of Linac triggers, given at
        mon_index=0.
    sample_trig : int
        The triggers used to calculate the count rate for the SAMPLE. 
        Typically this should be the number of Linac triggers, given at
        mon_index=0.
    
    Methods
    -------
    
    calc_count_rate
        Calculates the count rate for the sample given, such as OPEN or
        SAMPLE
    calc_open_rate
        Sets the 'cps' and 'dcps' columns in DataFrame open_data with the
        calculation from calc_count_rate
    calc_sample_rate
        Sets the 'cps' and 'dcps' columns in DataFrame sample_data with the
        calculation from calc_count_rate
    power_law 
        Uses coefficients given to calculate the power law function at the 
        time-of-flight values given
    double_exp 
        Uses coefficients given to calculate the double_exp function at the 
        time-of-flight values given
    calc_trans
        Sets the attribute 
    calc_sigma
        Calculates the cross section from transmission and atoms/barn thickness
        
    Notes
    -----
    Grouping for the transmission goes as follows:

    .. math:: C_g = \\sum_i^NC_i, \\qquad b_w = (N)(b_{w,raw})

    .. math:: \\dot C_{g} = \\frac{C_g}{b_w trig}

    .. math:: T_g = \\frac{\\dot C_{g,s}}{\\dot C_{g,O}} = \\frac{\\frac{\\sum_i^NC_s}{trig_sb_w}}{\\frac{\\sum_i^NC_O}{trig_Ob_w}}

    Background has four functions for now:
    
    `power_law`

    .. math:: a(tof)^{-b}+c

    `double_exp`

    .. math:: ae^{-b(tof)}+ce^{-d(tof)}+e

    `double_power_law`

    `exp_log`

    .. math:: f(t) = e^{ a + (b)t + c/ln(t) }

    Transmission from background subtracted count rates:

    .. math:: T_g = \\frac{\\dot C_{g,s}-\\dot B_{s}}{\\dot C_{g,O}-\\dot B_O}

    The error on transmission is propagated from the error on the counts (square root) and uses an 
    input constant fraction error on the background.

    .. math:: \\Delta T = \\sqrt{\\left(\\frac{\\partial T}{\\partial x}\\Delta x\\right)^2 + ...} 

    .. math:: = \\sqrt{\\left(\\frac{\\Delta \\dot C_s}{\\dot C_O - \\dot B_O}\\right)^2 + \\left(\\frac{\\Delta \\dot B_s}{\\dot C_O - \\dot B_O}\\right)^2 +\\left(\\frac{(\\dot C_s - \\dot B_s)\\Delta \\dot C_O}{\\dot C_O - \\dot B_O}\\right)^2 +\\left(\\frac{(\\dot C_s - \\dot B_s)\\Delta \\dot B_O}{\\dot C_O - \\dot B_O}\\right)^2}

    If pandas throws an "unsupported type" error, check your input file that is being
    read. It's possible you gave a file to be grouped, but you didn't tell the program
    to group.
    
    """
    def __init__(self,open_data_name,open_mon_name,sample_data_name,sample_mon_name,
                 datafile_names=None,tau=None,correct_deadtime=False,mon_index=0,
                 bkgfunction='power_law',o_bkgcoeff=None,s_bkgcoeff=None,factor=None,
                 c_pts=None,FP=None,samp_bkg_err=0.2,open_bkg_err=0.2,optstat_group=False,
                 D=None,res_per_bin=None,E_cpts=None,verbosity=False,binedge=False):
    
    
        self.mon_index = mon_index
        self.correct_deadtime = correct_deadtime
        self.bkgfunction = bkgfunction
        self.samp_bkg_err = samp_bkg_err
        self.open_bkg_err = open_bkg_err

        # ---------------------------------------------------------------------------------
        # ----- decide how to choose some attributes and defaults -------------------------
        # ---------------------------------------------------------------------------------
        if o_bkgcoeff is not None:
            self.o_bkgcoeff = o_bkgcoeff
        if s_bkgcoeff is not None:
            self.s_bkgcoeff = s_bkgcoeff
        
        if bkgfunction == 'power_law':
            if o_bkgcoeff is None:
                self.o_bkgcoeff = [8e5,1.4,102]
            if s_bkgcoeff is None:
                self.s_bkgcoeff = [8e5,1.45,83]
        elif bkgfunction == 'double_exp':
            if o_bkgcoeff is None:
                self.o_bkgcoeff = [4e5,2e-1,4e3,8e-3,120]
            if s_bkgcoeff is None:
                self.s_bkgcoeff = [4e5,2e-1,4e3,8e-3,120]
        elif bkgfunction == 'double_power_law':
            if o_bkgcoeff is None or s_bkgcoeff is None:
                raise ValueError("User must specify double_power_law bkg coefficients.")
        elif bkgfunction == 'exp_power_mix':
            if o_bkgcoeff is None or s_bkgcoeff is None:
                raise ValueError("User must specify exp_power_mix bkg coefficients.")
        elif bkgfunction == 'exp_log':
            if o_bkgcoeff is None or s_bkgcoeff is None:
                raise ValueError("User must specify exp_log bkg coefficients.")

        # the names for the Ogrp function from rpixdr (tof,bin_width,counts)
        if datafile_names is None:
            self.datafile_names = ['tof','bin_width','counts']
        else:
            self.datafile_names = datafile_names

        if E_cpts is None:
            E_cpts = [1e2,1e3,1e4]
        
        # ---------------------------------------------------------------------------------
        # ----- read in the counts, triggers, and monitors --------------------------------
        # ---------------------------------------------------------------------------------
        
        grouping_is_used = False
        # read in the counts data
        if factor is None and c_pts is None and optstat_group==False:
            # -----------------------------------------------------------------
            # ---- read in optgroup files that have already been grouped ------
            # -----------------------------------------------------------------
            self.open_data   = pd.read_csv(open_data_name  ,delim_whitespace=True,names=self.datafile_names)
            self.sample_data = pd.read_csv(sample_data_name,delim_whitespace=True,names=self.datafile_names)
        elif optstat_group == True:
            # -----------------------------------------------------------------
            # ---- group by "optimum statistics" ------------------------------
            # -----------------------------------------------------------------
            if D==None or res_per_bin==None or FP==None:
                print("Fix input parameters.")
                raise ValueError("Missing: level spacing (D) or res. desired per bin (res_per_bin) or FP.")
            # ---- reset the column names -------------------------------------------------
            self.datafile_names = ['ch','counts','dcounts']
            # ---- read in the fixed bin width data ---------------------------------------
            ug_open_data = pd.read_csv(open_data_name,delim_whitespace=True,names=self.datafile_names,
                                         skiprows=14)
            ug_samp_data = pd.read_csv(sample_data_name,delim_whitespace=True,names=self.datafile_names,
                                         skiprows=14)
            # ---- read in the bin width --------------------------------------------------
            open_bw = np.genfromtxt(open_data_name  ,unpack=True,skip_header=12,max_rows=1)
            samp_bw = np.genfromtxt(sample_data_name,unpack=True,skip_header=12,max_rows=1)
            # ---- create the open and sample data ----------------------------------------
            self.open_data   = pd.DataFrame(tt.optstat_group([ug_samp_data.ch*float(open_bw[0]),ug_open_data.counts],
                               FP,D,res_per_bin,E_cpts=E_cpts,verbose=verbosity,binedge=binedge).T,
                               columns=['tof','bin_width','counts','dcounts','tof_low'])
            self.sample_data = pd.DataFrame(tt.optstat_group([ug_samp_data.ch*float(samp_bw[0]),ug_samp_data.counts],
                               FP,D,res_per_bin,E_cpts=E_cpts,verbose=verbosity,binedge=binedge).T,
                               columns=['tof','bin_width','counts','dcounts','tof_low'])
        elif factor is not None and c_pts is not None and FP is not None:
            # -----------------------------------------------------------------
            # ---- Group with compression point grouping ----------------------
            # -----------------------------------------------------------------
            grouping_is_used = True
            # ---- reset the column names -------------------------------------------------
            self.datafile_names = ['ch','counts','dcounts']
            # ---- read in the fixed bin width data ---------------------------------------
            ug_open_data   = pd.read_csv(open_data_name,delim_whitespace=True,names=self.datafile_names,
                                         skiprows=14)
            ug_samp_data = pd.read_csv(sample_data_name,delim_whitespace=True,names=self.datafile_names,
                                         skiprows=14)
            # ---- read in the bin width --------------------------------------------------
            open_bw = np.genfromtxt(open_data_name  ,unpack=True,skip_header=12,max_rows=1)
            samp_bw = np.genfromtxt(sample_data_name,unpack=True,skip_header=12,max_rows=1)
            # ---- create the open and sample data ----------------------------------------
            self.open_data   = pd.DataFrame(tt.comp_group([ug_samp_data.ch*float(open_bw[0]),ug_open_data.counts],
                               factor,c_pts,FP,binedge=binedge).T,columns=['tof','bin_width','counts','dcounts'])
            self.sample_data = pd.DataFrame(tt.comp_group([ug_samp_data.ch*float(samp_bw[0]),ug_samp_data.counts],
                               factor,c_pts,FP,binedge=binedge).T,columns=['tof','bin_width','counts','dcounts'])
        else:
            err_str = "Data could not be read. If grouping is desired define factor, c_pts, and FP."
            print('If grouping, NORSUM file, constant bin width must be used.')
            raise ValueError(err_str)
        
        # read in the monitors and triggers
        self.open_mon   = pd.read_csv(open_mon_name  ,names=['mon'])
        self.sample_mon = pd.read_csv(sample_mon_name,names=['mon'])
        
        # ---------------------------------------------------------------------------------
        # ----- test data that was read just read -----------------------------------------
        # ---------------------------------------------------------------------------------
        # check if data is a 2-d array
        if (len(self.open_data.shape) > 2) or (len(self.sample_data.shape) > 2):
            raise ValueError("Data should be 2-d: columns and rows")
        # check if data has 3 columns (this is true for Ogrp data)
        if (self.open_data.shape[1] != 3) or (self.sample_data.shape[1] != 3):
            if factor == None and optstat_group == False:
                raise ValueError("Data should only have 3 columns")
        # check if user gave Ogrp files but also gave grouping input
        if grouping_is_used and 'Ogr' in open_data_name:
            raise ValueError('Name mismatch for grouping')
        # check if monitor files are 11 monitors
        # -- consider revising this requirement.
        if (len(self.open_mon) > 11) or (len(self.sample_mon) > 11):
            raise ValueError("There should be 11 monitors in the mon file")
        # check if there are extra columns in the monitor files
        if (self.open_mon.shape[1] > 1) or (self.sample_mon.shape[1] > 1):
            raise ValueError("The monitor file should be one column of 10 data")
            
        # if triggers is not specified by user, set it to first element of monitor array
        self.sample_trig = self.sample_mon['mon'][self.mon_index] 
        self.open_trig = self.open_mon['mon'][self.mon_index]
        
        # ---------------------------------------------------------------------------------
        # ----- convert the counts to count rate for each ---------------------------------
        # ---------------------------------------------------------------------------------
        self.calc_open_rate()
        self.calc_sample_rate()
        
        # ---------------------------------------------------------------------------------
        # ----- calculate bkg function along the tof grid ---------------------------------
        # ---------------------------------------------------------------------------------
        if bkgfunction == 'double_exp':
            self.open_data['bkg'] = self.double_exp(self.open_data.tof,self.o_bkgcoeff[0],
                                                    self.o_bkgcoeff[1],self.o_bkgcoeff[2],
                                                    self.o_bkgcoeff[3],self.o_bkgcoeff[4])
            self.sample_data['bkg'] = self.double_exp(self.sample_data.tof,self.s_bkgcoeff[0],
                                                        self.s_bkgcoeff[1],self.s_bkgcoeff[2],
                                                        self.s_bkgcoeff[3],self.s_bkgcoeff[4])
        elif bkgfunction == 'power_law':
            self.open_data['bkg'] = self.power_law(self.open_data.tof,self.o_bkgcoeff[0],
                                                   self.o_bkgcoeff[1],self.o_bkgcoeff[2])
            self.sample_data['bkg'] = self.power_law(self.sample_data.tof,self.s_bkgcoeff[0],
                                                       self.s_bkgcoeff[1],self.s_bkgcoeff[2])
        elif bkgfunction == 'double_power_law':
            self.open_data['bkg'] = self.double_power_law(self.open_data.tof,FP,
                                                    self.o_bkgcoeff[0],
                                                    self.o_bkgcoeff[1],self.o_bkgcoeff[2],
                                                    self.o_bkgcoeff[3],self.o_bkgcoeff[4])
            self.sample_data['bkg'] = self.double_power_law(self.sample_data.tof,FP,
                                                    self.s_bkgcoeff[0],
                                                    self.s_bkgcoeff[1],self.s_bkgcoeff[2],
                                                    self.s_bkgcoeff[3],self.s_bkgcoeff[4])
        elif bkgfunction == 'exp_power_mix':
            self.open_data['bkg'] = self.exp_power_mix(self.open_data.tof,self.o_bkgcoeff[0],
                                                    self.o_bkgcoeff[1],self.o_bkgcoeff[2],
                                                    self.o_bkgcoeff[3],self.o_bkgcoeff[4])
            self.sample_data['bkg'] = self.exp_power_mix(self.sample_data.tof,self.s_bkgcoeff[0],
                                                        self.s_bkgcoeff[1],self.s_bkgcoeff[2],
                                                        self.s_bkgcoeff[3],self.s_bkgcoeff[4])
        elif bkgfunction == 'exp_log':
            self.open_data['bkg'] = self.exp_log(self.open_data.tof,self.o_bkgcoeff[0],
                                                 self.o_bkgcoeff[1],self.o_bkgcoeff[2],
                                                 self.o_bkgcoeff[3])
            self.sample_data['bkg'] = self.exp_log(self.sample_data.tof,self.s_bkgcoeff[0],
                                                   self.s_bkgcoeff[1],self.s_bkgcoeff[2],
                                                   self.s_bkgcoeff[3])
        else:
            raise ValueError('Options for bkgfunction are: \'double_exp\',\'power_law\',',
                '\'double_power_law\',\'exp_power_mix\'')
        
        # ---------------------------------------------------------------------------------
        # -----calculate transmission -----------------------------------------------------
        # ---------------------------------------------------------------------------------
        self.calc_trans()

        # ---------------------------------------------------------------------------------
        # ----- finished with initialization ----------------------------------------------
        # ---------------------------------------------------------------------------------
        
    def calc_count_rate(self,COUNTS,BIN_WIDTH,TRIGGERS):
        """"
        Function to convert counts to count rate.
        
        Parameters
        ----------
        
        counts : array-like
            1-d vector of counts that were collected for each bin
        bin_width : array-like
            1-d vector of widths in micro-seconds for each bin associated
            with the counts in the counts vector
        triggers : float
        The number of times that the bin was "open" or collecting 
            counts data
        correct_deadtime : bool, optional
            If true, the counts data will be corrected for the dead time
            of the detection system.
        tau : float, optional
            Tau is the dead time of the data collection system in micro-
            seconds. If correct_deadtime is true, tau is required to correct
            for the dead time of the system
        
        Returns
        -------
        
        cps : array-like
            1-d vector of count rate in counts per second
        dcps : array-like
            1-d vector of the error in the count rate

        Notes
        -----
        .. math:: CPS = \\frac{C}{b_w\\cdot trig}
        
        """
        
        if self.correct_deadtime == True:
            if tau == None:
                raise ValueError("If dead time correction is wanted, specify tau.")
            # TODO : write the dead time correction
            else:
                raise ValueError("Doh! Dead time has not been implemented yet.")
            # correct counts for dead time
            
            # calculate count rate
        
            # calculate error on the count rate
            
        else:
            # calculate count rate
            cps = COUNTS/BIN_WIDTH/TRIGGERS/1e-06
            
            # calculate error on the count rate
            dcps = np.sqrt(COUNTS)/BIN_WIDTH/TRIGGERS/1e-06
        
            return cps,dcps
                
    def calc_open_rate(self):
        """
        Call calc_count_rate for the open_data
        """
        self.open_data['cps'],self.open_data['dcps'] = self.calc_count_rate(self.open_data.counts,
                                                self.open_data.bin_width,self.open_trig)
    def calc_sample_rate(self):
        """
        Call calc_count_rate for the sample_data
        """
        self.sample_data['cps'],self.sample_data['dcps'] = self.calc_count_rate(self.sample_data['counts'],
                                                self.sample_data['bin_width'],self.sample_trig)
        
        
        
    # ---------------------------------------------------------------------------------
    # ----- determine a bkg function for open and sample ------------------------------
    # ---------------------------------------------------------------------------------
    # --> for now I'll have a couple functions to adjust on the fly
        
    def double_exp(self,t,a,b,c,d,e):
        """
        .. math:: f(t) = (a)e^{-t(b)} + (c)e^{-t(d)} + e
        """
        return a*np.exp(-t*b) + c*np.exp(-t*d) + e
        
    def power_law(self,t,a,b,c):
        """
        .. math:: f(t) = (a)t^{-(b)} + c
        """
        return a*t**(-b) + c

    def double_power_law(self,t,FP,a1,b1,a2,b2,c):
        """
        .. math:: f(t) = (a1)t^{(b1)} + (a2)\\left(\\frac{72.3FP}{t}\\right)^{2(b2)} + c
        """
        return a1*t**b1 + a2*tt.tofe(t,FP)**b2 + c

    def exp_power_mix(self,t,a,b,c,d,e):
        """
        .. math:: f(t) = (a)e^{-t(b)} + (c)t^{-(d)} + e
        """
        return a*np.exp(-t*b) + c*t**(-d) + e

    def exp_log(self,t,a,b,c,d):
        """
        .. math:: f(t) = e^{ a + (b)t + c/ln(t) } + d
        """
        return np.exp(a + b*t + c/np.log(t)) + d
        
    # ---------------------------------------------------------------------------------
    # ----- calculate the transmission ------------------------------------------------
    # ---------------------------------------------------------------------------------
    def calc_trans(self):
        """
        .. math:: T = \\frac{\\dot C_s - \\dot B_s}{\\dot C_o - \\dot B_o}
        """
        self.trans = ((self.sample_data.cps - self.sample_data.bkg)/
                             (self.open_data.cps - self.open_data.bkg))

        self.dtrans = np.sqrt( 
            (self.sample_data.dcps/(self.open_data.cps-self.open_data.bkg))**2 +

            (self.samp_bkg_err*self.sample_data.bkg/(self.open_data.cps-self.open_data.bkg))**2 +
            
            ((self.sample_data.bkg-self.sample_data.cps)/(self.open_data.cps-self.open_data.bkg)**2 * 
                self.open_data.dcps)**2 +

            ((self.sample_data.cps-self.sample_data.bkg)/(self.open_data.cps-self.open_data.bkg)**2 * 
                self.open_data.bkg*self.open_bkg_err)**2
                              )

    def calc_sigma(self,N):
        """
        .. math:: \\sigma_t = \\frac{-1}{N}log(T) \\pm \\frac{\\Delta T}{NT}
        """
        # cross section for this samples transmission
        self.sig = -1/N*np.log(self.trans)
        # error on the cross section for the sample transmission
        self.dsig = self.dtrans/N/self.trans
        

def calc_cov(mon_list,dmon_list,room_bkgs,droom_bkgs,norm_list,dnorm_list,tof,sample_cps,
             sample_dcps,open_cps,open_dcps,bkg_function,bkg_pars,bkg_pars_cov,split=False,
             trans_idc_output=False):
    """
    Calculate covariance for transmission T = ( a1*Cs - a2*ks*B - B0s )/( a2*Co - a4*ko*B - B0o )

    To avoid code complexity, for now I'll force the user to specify everything up front instead
    of passing a Trans class. Available background functions: "power_law", "exp"

    Parameters
    ----------
    mon_list : list
        A list of the monitor values in order: [a1,a2,a3,a4]
    dmon_list : list
        A list of the uncertainty on monitor values in matching order
    room_bkgs : list
        A list of the room/ambient backgrounds a.k.a. B-zeros in order: [B0s,B0o]
    droom_bkgs : list
        A list of the uncertainty on room/ambient backgrounds in matching order
    norm_list : list
        A list of the normalizations for background in order: [ks,ko]
    dnorm_list : list
        A list of the uncertainty on normalizations for backgrounds in matching order
    tof : array-like
        A 1-dim array of time-of-flight for sample and open count rates. This must
        be "true" TOF, i.e. time-zero subtracted in order to get the background 
        correct.
    sample_cps : array-like
        A 1-dim array of sample count rate in order of time-of-flight (Cs)
    sample_dcps : array-like
        A 1-dim array of unc. on sample count rate in order of time-of-flight (dCs)
    open_cps : array-like
        A 1-dim array of open count rate in order of time-of-flight (Co)
    open_dcps : array-like
        A 1-dim array of unc. on open count rate in order of time-of-flight (Co)
    bkg_function : str
        The name of the background function used, matching the names in class Trans 
    bkg_pars : list
        The parameters needed for the background function (B-zero subtracted!)
    bkg_pars_cov : array-like
        A 2-dim array of the covariance for the fitted parameters.
    split : bool, optional
        Whether to return two parts of covariance instead of the sum, the statistical
        and the systematic
    trans_idc_output : bool, optional
        Extra output to be used with function `write_sammy_idc_file()`. In addition to
        default output `cov`, 3 extra variables are returned: `stat`, `syst_full`, and
        `sys_der_list`. 
    
    Returns
    -------
    cov : array-like
        A 2-dim numpy array with full covariance
    stat,syst : tuple
        A tuple of 2 2-dim arrays, the statistical and systematic covariance

    """
    available_bkg_funcs = np.array(['power_law','exp'])
    if not (available_bkg_funcs == bkg_function).any():
        raise ValueError("background function not available")
    # ----------------------------
    # Rename to simplify
    # ----------------------------
    ks,ko,dks,dko = norm_list[0],norm_list[1],dnorm_list[0],dnorm_list[1]
    a1,a2,a3,a4 = mon_list[0],mon_list[1],mon_list[2],mon_list[3]
    da1,da2,da3,da4 = dmon_list[0],dmon_list[1],dmon_list[2],dmon_list[3]
    b0s,b0o,db0s,db0o = room_bkgs[0],room_bkgs[1],droom_bkgs[0],droom_bkgs[1]
    cs,co,dcs,dco = sample_cps,open_cps,sample_dcps,open_dcps
    N,D,bfit = 1,1,0
    # ----------------------------
    # Bkg specific variables 
    # ----------------------------
    if bkg_function == 'power_law':
        A,B = bkg_pars[0],bkg_pars[1]
        bfit = A*tof**(-B)
        N = (a1*cs-a2*ks*bfit-b0s)
        D = (a3*co-a4*ko*bfit-b0o)
        dda = (-D*ks + N*ko)*tof**-B/D**2 
        ddb = (-D*ks + N*ko)*-1*bfit*np.log(tof)/D**2
        fit_ders = [dda,ddb]
    if bkg_function == 'exp':
        A,B = bkg_pars[0],bkg_pars[1]
        bfit = A*np.exp(-tof*B)
        N = (a1*cs-a2*ks*bfit-b0s)
        D = (a3*co-a4*ko*bfit-b0o)
        dda  = -1*(ks*D+ko*N)/(A*D**2)
        ddb  = (ks*D+ko*N)*bfit*tof/D**2
        fit_ders = [dda,ddb]
    # ----------------------------
    # Bkg agnostic variables
    # ----------------------------
    # statistical
    ddcs = a1/D
    ddco = -a3*N/D**2
    # systematic
    ddks = -a2*bfit/D
    ddko = a4*N*bfit/D**2
    ddb0s = -1/D
    ddb0o = 1*N/D**2
    dda1 = cs/D
    dda2 = -ks*bfit/D
    dda3 = -co*N/D**2
    dda4 = ko*bfit*N/D**2
    # statistical err and derivs *order matters!*
    stat_err_list = [dcs,dco]
    stat_der_list = [ddcs,ddco]
    # systematic err and derivs *order matters!*
    sys_err_list = [dks,dko,db0s,db0o,da1,da2,da3,da4] # doesn't inc. A,B they come with covar mat
    sys_der_list = np.concatenate([fit_ders,[ddks,ddko,ddb0s,ddb0o,dda1,dda2,dda3,dda4]])
    
    stat = mt.stat_cov(stat_err_list,stat_der_list)
    syst = mt.sys_cov(sys_err_list,sys_der_list,cov_mat=bkg_pars_cov) # Note bkg. cov. matrix here
    
    cov = stat+syst

    # ----------------------------
    # Return options
    # ----------------------------
    if trans_idc_output:
        # Form the full syst par cov 
        K = len(bkg_pars_cov)+len(sys_err_list)
        syst_full = np.zeros((K,K))
        syst_full[0:len(bkg_pars_cov),0:len(bkg_pars_cov)] = bkg_pars_cov
        for i in range(len(sys_err_list)):
            index = i+len(bkg_pars_cov)
            syst_full[index,index] = sys_err_list[i]**2

        return cov,stat,syst_full,sys_der_list
    if split:
        return stat,syst
    else:
        return cov

def write_covar_file(filename,energy,t,dt,sys_err_list,stat_err_list,sys_der_list,stat_der_list,
                 sys_err_str_list,ptwise_str_list,sys_cov=None,sys_cov_str_list=None,
                 high_precision=False):
    """
    Write a file with the information necessary to calculate the point by point covariance
    of a function.

    Parameters
    ----------
    filename : str
        The full file path to the file that is being written to
    energy : array
        The 1-d vector of energy. This should be the same length as all
        the other pointwise quantities given.
    t : array-like
        The observable, in this case transmission. Should be the same length as
        energy
    dt : array-like
        The uncertainty on the observable (for convenience). Should be the same
        length as energy, and should be the diagonal of the covariance matrix.
    sys_err_list : array-like
        The list of systematic errors for the function. These are constant 
        along the length of the function so are only single floats.
    stat_err_list : array-like
        A 2-d array where (len(stat_err_list) = no. of statistical variables),
        and (len(stat_err_list[0]) = len(energy)). This is a list of vectors 
        of the pointwise statistical error for each statistical variable.
    sys_der_list : array-like
        A list very similar to stat_err_list, but the number of vectors 
        corresponds to the number of systematic variables in the function.
        Each vector in this list is the derivative of the function with 
        respect to that variable. **The order of the derivative lists
        (systematic and statistical) should match the order of the error
        lists, in terms of which variables are listed**. 
    stat_der_list : array-like
        A list very similar to sys_der_list, but contains derivatives of 
        the function with respect to the statistical variables. It bears
        repeating that  **The order of the derivative lists
        (systematic and statistical) should match the order of the error
        lists, in terms of which variables are listed**.
    sys_err_str_list : list
        A list of strings in the order of the given systematic errors. This
        is used to label the information in the file.
    ptwise_str_list : list
        A list of strings in the order of [stat_err_list],[sys_der_list],
        [stat_der_list] that will be used to label the columns of data 
        listed in the file.
    sys_cov : array-like, optional
        If there are correlated systematic input parameters (e.g. parameters
        in the fitted function for background in transmission) the user can 
        provide the covariance matrix here. The covariance matrix is put at 
        the first rows and columns of the input systematic covariance matrix
        so **care must be taken to ensure the order of the derivatives matches
        the new order of errors given the inserted variables in sys_cov**. As
        an example if i have [dcs] as sys_err_list and [[da**2,dadb],[dbda,db**2]]
        as the sys_cov, my sys_der_list will look like [dda,ddb,ddcs].
    sys_cov_str_list : list, optional
        If the sys_cov is given, label the columns in the file with this list. 
        Optional unless the sys_cov variable is given.
    high_precision : bool, optional
        If the precision written to the file is too low (default 1e-6), write
        the floats with higher precision (1e-12).

    Returns
    -------
    none : ascii file
        Returns a somewhat annotated file with all the information necessary to
        reproduce the point-by-point covariance of the function. This output file
        is readable by read_covar_file().


    """
    def sixe(x):
        return '{:10.6e}'.format(x)
    if high_precision:
        def sixe(x):
            return '{:10.12e}'.format(x)

    with open(filename,'w+') as f:
        f.write('######\n# Systematic errors\n')
        f.write('# ')
        for i,string in enumerate(sys_err_str_list):
            f.write(string)
            if i != len(sys_err_str_list)-1:
                f.write(',')
        f.write('\n######\n')
        for i,err in enumerate(sys_err_list):
            f.write(sixe(err))
            if i != len(sys_err_list)-1:
                f.write(',')
        f.write('\n')
        if sys_cov is not None:
            f.write('######\n# Correlated syst. errors\n')
            f.write('# ')
            for i,string in enumerate(sys_cov_str_list):
                f.write(string)
                if i != len(sys_cov_str_list)-1:
                    f.write(',')
            f.write('\n######\n')
            for i,row in enumerate(sys_cov):
                for j,elem in enumerate(row):
                    f.write(sixe(elem))
                    if j != len(row)-1:
                        f.write(',')
                f.write('\n')
        f.write('######\n# E,t,dt,[stat. error],[sys derivatives],[stat derivatives]\n')
        f.write('# ')
        for i,string in enumerate(ptwise_str_list):
            f.write(string)
            if i != len(ptwise_str_list)-1:
                f.write(',')
        f.write('\n######\n')
        for i in range(len(sys_der_list[0])):
            f.write(sixe(energy[i])+','+sixe(t[i])+','+sixe(dt[i])+',')
            for j in range(len(stat_err_list)):
                f.write(sixe(stat_err_list[j][i])+',')
            for j in range(len(sys_der_list)):
                f.write(sixe(sys_der_list[j][i])+',')
            for j in range(len(stat_der_list)):
                f.write(sixe(stat_der_list[j][i]))
                if j != len(stat_der_list)-1:
                    f.write(',')
            f.write('\n')

def read_covar_file(filename,stat_err_cols,sys_der_cols,stat_der_cols,corr_sys_input=0):
    """
    Read a covariance file written by the write_covar_file() function in nuctools

    Parameters
    ----------
    filename : str
        The full file path to the file being read
    stat_err_cols : int
        The number of columns of statistical error in the file. Should be the same as 
        the number of statistical variables in the fucntion.
    sys_der_cols : int
        The number of columns of systematic derivatives in the file. Should be the same
        as the number of systematic variables in the fucntion.
    stat_der_cols : int
        The number of columns of statistical derivatives in the file. Should be the 
        same as the number of statistical variables in the fucntion.
    corr_sys_input : int, optional
        The nubmer of correlated variables given in a covariance matrix for systematic
        variables. Default is 0, meaning there is no covariance matrix in the file.

    Returns
    -------
    cy : numpy array
        The point-by-point covariance of the function

    """

    # read in the systematic error list
    sys_err_list = np.genfromtxt(filename,unpack=True,skip_header=4,delimiter=',',max_rows=1)
    # is there a matrix for correlated systematic error inputs?
    if corr_sys_input > 0:
        sys_cov = np.genfromtxt(filename,skip_header=9,delimiter=',',max_rows=corr_sys_input)
    # read in the point-wise data for statistical error and systematic/statistical derivatives
    ptwise = np.genfromtxt(filename,skip_header=9+corr_sys_input+4,delimiter=',',unpack=True)
    
    stat_err_list = ptwise[3:3+stat_err_cols,:]
    sys_der_list = ptwise[3+stat_err_cols:3+stat_err_cols+sys_der_cols,:]
    stat_der_list = ptwise[3+stat_err_cols+sys_der_cols:3+stat_err_cols+sys_der_cols+stat_der_cols,:]
    
    stat = mt.stat_cov(stat_err_list,stat_der_list)
    if corr_sys_input > 0:
        syst = mt.sys_cov(sys_err_list,sys_der_list,cov_mat=sys_cov)
    else:
        syst = mt.sys_cov(sys_err_list,sys_der_list)

    E = ptwise[0]
    observable = ptwise[1]
    unc_observable = ptwise[2]
    
    cy = stat+syst
    
    return E,observable,unc_observable,cy







import numpy as np
import pandas as pd
from . import tof_tools as tt

__all__ = ['ags','e1p0','e2p0','e3p0']

class ags:
    """
    Python class to contain AGS-like functions and information. Please see the AGS Manual 
    (B. Becker, C. Bastian, J. Heyse, S. Kopecky, P. Schillebeeckx, AGS - Analysis of Geel
    Spectra, European Commission, Joint Research Centre, September 2014) for more detail: 
    https://www.oecd-nea.org/jcms/pl_19568/ags-analysis-of-geel-spectra-users-manual

    Attributes
    ----------
    tof : array-like
        A 1-d array that is a Pandas Series of time-of-flight values,
        typically given in [us]
    counts : array-like
        A 1-d array that is a Pandas Series of binned counts
    cps : array-like
        A 1-d array that is a Pandas Series of count rate given in 
        counts per second
    energy : array-like
        A 1-d array that is a Pandas Series of energy values transformed 
        from the time-of-flight values

    """
    def __init__(self):
        self.tof = None
        self.dtof = None
        self.counts = None
        self.cps = None
        self.energy = None
        self.obs = None
        self.unc_obs = None
        
    def read_grouped_counts(self,spectrum,comp_pt,comp_fct,binsize,triggers):
        """
        Read in the grouped counts file from AGL and populate the 
        counts, tof, and cps attributes of the ags class
        
        Parameters
        ----------
        filename : str
            The full file path to the AGL grouped counts file
        comp_pt : array-like
            Compression points given in bin numbers (integers). Must
            be of length == len(comp_fct)+1
        comp_fct : array-like
            Compression factors for each group. These are specified in 
            a separate text file from AGL. Typically in powers of 2 
            (2^N).
        binsize : float
            The width of the base bin in [us]
        triggers : integers
            the number of times the linac fired

        Returns
        -------
        nothing : None
            Populates the attributes of the class: tof, counts, cps

        Examples
        --------

        Notes
        -----

        """
        cp = np.array(comp_pt)
        cf = 2**np.array(comp_fct)
        #hist = pd.read_csv(filename,delim_whitespace=True,names=['bin','counts'])
        hist = pd.DataFrame({
            'counts'    : spectrum,
            'dtof' : np.zeros(len(spectrum))
            })
        hist['cps'],hist['dcps'],hist['tof'] = 0,0,0
        
        for i in range(len(cp)-1):
            hist.loc[cp[i]:cp[i+1]-1,'cps'] = hist.counts[cp[i]:cp[i+1]]/(cf[i]*triggers*binsize*1e-6)
            hist.loc[cp[i]:cp[i+1]-1,'dcps'] = np.sqrt(hist.counts[cp[i]:cp[i+1]])/(cf[i]*triggers*binsize*1e-6)
            hist.loc[cp[i]:cp[i+1]-1,'dtof'] = cf[i]*binsize
            if i==0:
                hist.loc[cp[i]:cp[i+1]-1,'tof'] = 0+np.arange(cp[i+1]-cp[i])*binsize*cf[i]
            else:
                hist.loc[cp[i]:cp[i+1]-1,'tof'] = hist.tof[cp[i]-1]+np.arange(cp[i+1]-cp[i])*binsize*cf[i]
                
        self.tof = hist.tof
        self.dtof = hist.dtof
        self.cps = hist.cps
        self.dcps = hist.dcps
        self.counts = hist.counts

    def calc_energy(self,FP,t0):
        """
        Calculate energy from the existing TOF spectrum

        Parameters
        ----------
        FP : float
            The flight path length the neutron traveled [m]
        t0 : float
            The time-zero used to correct the TOF spectrum [us]
        """
        self.energy = tt.tofe(self.tof-t0,FP)

def e1p0(tof,p1,p2,p3):
    """
    Background function for TOF spectra

    Parameters
    ----------
    tof : array-like
        The time-of-flight spectrum
    p1 : float
        constant background
    p2 : float
        multiplier on 1st exponential
    p3 : float
        multiplier on time-of-flight in 1st exponent
    p4 : float
        constant added to 1st exponent

    Returns
    -------
    e1p0 : array-like
        The function in the length of t (see notes)

    Notes
    -----
    .. math:: f(t) = p1 + p2e^{p3t+p4}
    """
    return p1 + p2*np.exp(p3*tof)

def e2p0(tof,p1,p2,p3,p5,p6):
    """
    Background function for TOF spectra

    Background function for TOF spectra

    Parameters
    ----------
    tof : array-like
        The time-of-flight spectrum
    p1 : float
        constant background
    p2 : float
        multiplier on 1st exponential
    p3 : float
        multiplier on time-of-flight in 1st exponent
    p4 : float
        constant added to 1st exponent
    p5-p7 : float
        (see equation in notes)

    Returns
    -------
    e2p0 : array-like
        The function in the length of t (see notes)

    Notes
    -----
    .. math:: f(t) = p1 + p2e^{p3t+p4} + p5e^{p6t+p7}
    """
    return p1 + p2*np.exp(p3*tof) + p5*np.exp(p6*tof)

def e3p0(tof,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10):
    """
    Background function for TOF spectra

    Parameters
    ----------
    tof : array-like
        The time-of-flight spectrum
    p1 : float
        constant background
    p2 : float
        multiplier on 1st exponential
    p3 : float
        multiplier on time-of-flight in 1st exponent
    p4 : float
        constant added to 1st exponent
    p5-p10 : float
        (see equation in notes)

    Returns
    -------
    e3p0 : array-like
        The function in the length of t (see notes)

    Notes
    -----
    .. math:: f(t) = p1 + p2e^{p3t+p4} + p5e^{p6t+p7} + p8e^{p9t+p10}
    """
    return p1 + p2*np.exp(p3*tof+p4) + p5*np.exp(p6*tof+p7) + p8*np.exp(p9*tof+p10)


def dtco(dtof,counts,dead_time,trigs):

    lc = len(counts)
    SUM = np.zeros(lc)
    time_window = 0.0
    for i in range(lc):
        if tof[i] < dead_time:
            SUM[i] = 0.0
        else: 
            i0 = None
            time_sum = 0
            tind = i-1
            while time_sum < dead_time:
                time_sum += dtof[tind]
                tind-1
            SUM = 1.
    corr_counts = trigs * ( -np.log( 1-(counts/trigs)/(1-SUM) ) )









import re,os,json

__all__ = ['get_protons','get_single_elem_list']

def get_protons(chem_symbol):
    """
    """
    f_protons = os.path.join(os.path.dirname(__file__),'data','protons.json')
    with open(f_protons,'r') as f:
        protons = json.load(f)
    
    return protons[chem_symbol]

def get_single_elem_list(unary_salts):
    """
    Assuming only chloride salts, break list of unary salt 
    mixtures into a single list of unique chemical elements

    Parameters
    ----------
    unary_salts : list
        List of strings that describe formula for unary salts.

    Returns
    -------
    unique_elements : list
        A list of strings that are the unique elements for all
        salts listed 

    Notes
    -----
    Second author: Tarek Ghaddar

    """
    unique_elements = []
    for salt in unary_salts:
        elem_list = re.findall('[A-Z][^A-Z]*',salt)
        #print(elem_list)
        for elem in elem_list:
            # Getting only the letters from the string
            elem_letters = re.sub('[^a-zA-Z]+', '',elem)
            if elem_letters not in unique_elements:
                unique_elements.append(elem_letters)

    return unique_elements






import pandas as pd
import numpy as np
import h5py as h5
from . import tof_tools as tt
from . import math_tools as mt
from . import funky
import time

__all__ = ['read_h5rpcm','read_h5xscm','read_pds','write_sammy_idc_file',
           'print_chann','create_sam_inp','fmt_twenty','fmt_ten','fmt_par',
           'switch_par_flags','describe_norsum_header','LPTtoCOV','mini_cov']


def read_h5rpcm(filename):
    """
    New SAMMY covariance file is in HDF5 in the form of an upper 
    triangular matrix. Return covariance, parameters, etc.

    Parameters
    ----------
    filename : str
        The path and filename for the SAMMY covariance file in HDF5 format

    Returns
    -------
    ucov,upar,covind,ispup : tuple
        A tuple of numpy arrrays with the reduced parameters covariance matrix, 
        reduced parameters, indices for parameters into the covariance matrix, 
        and a list of 0,1 booleans on whether the parameter was PUP'd
    """
    with h5.File(filename,"r+") as f:
        flatcov = np.array(f['covar'])
        upar = np.array(f['param'])
        covind = np.array(f['covind'])
        try: 
            ispup = np.array(f['ispup'])
        except:
            ispup = None

    lp = len(upar)
    ucov = np.zeros((lp,lp))
    k = 0
    for i in range(lp):
        for j in range(i,lp):
            ucov[i,j] = flatcov[k]
            ucov[j,i] = flatcov[k]
            k+=1

    if ~np.all(ucov == ucov.T):
        raise ValueError("Covariance matrix is not symmetric!")

    return ucov,upar,covind,ispup


def read_h5xscm(filename):
    """
    New SAMMY XS covariance file is in HDF5 in the form of a lower
    triangular matrix (different than RPCM above!). Return cross section
    covariance

    Parameters
    ----------
    filename : str
        The name of the full file path to the H5 covariance file
    Returns
    -------
    xscov : array-like
        The 2-d symmetric array of covariance on the theoretical observable
        in SAMMY.
    """

    with h5.File(filename,"r+") as f:
        try: 
            flatxscov = np.array(f['xs_cov'])
        except:
            flatxscov = None
    
    lxc = len(flatxscov)
    lp = int(np.floor(np.sqrt(1+8*lxc)-1)/2)
    xscov = np.zeros((lp,lp))
    k = 0
    for i in range(lp):
        for j in range(0,i+1):
            xscov[i,j] = flatxscov[k]
            xscov[j,i] = flatxscov[k]
            k+=1

    if ~np.all(xscov == xscov.T):
        raise ValueError("Covariance matrix is not symmetric!")

    return xscov

def read_pds(pds_file):
    """
    Read the PDS partial derivative file and load it into a pandas dataframe. 

    Parameters
    ----------
    pds_file : str
        The path to the PDS file

    Returns
    -------
    df : DataFrame
        A pandas data frame, with columns: 
        "expdata","expunc","theory","dtdp1","dtdp2","dtdp3","dtdp{}",...
        all the way to the number of parameters used to create the 
        cross section
    """
    with open(pds_file,"r") as f:
        lines = f.readlines()
    evenlinesplit = []
    deriv_table = []
    pandas_cols = ["expdata","expunc","theory","dtdp1","dtdp2","dtdp3"]
    for i,line in enumerate(lines):
        if i==0:
            # first line
            numpars = int(line)
            print("Number of pars = ",numpars)
            for parnumber in range(4,numpars+1):
                pandas_cols.append("dtdp{}".format(parnumber))
            print(pandas_cols)
            continue
        if i==1:
            # second line
            reduced_pars = np.array(line.split()).astype(float)
            print("Reduced pars = ",reduced_pars)
            continue
        if (i-2)%2 == 0:
            # odd lines (as shown in SAMMY manual, but even when 0-indexed)
            evenlinesplit = line.split()
        if (i-2)%2 == 1:
            # even lines (as shown in SAMMY manual, but odd when 0-indexed)
            oddlinesplit = line.split()
            evenlinesplit.extend(oddlinesplit)
            newrow = np.array(evenlinesplit).astype(float)
            deriv_table.append(newrow)
    df = pd.DataFrame(deriv_table,columns=pandas_cols)
    return df

def write_sammy_idc_file(filename_abbrev,energy,obs,obserr,stat,syst,sys_der_list,high_precision=False):
    """
    Write a file with the information requested by the SAMMY code for calculating covariance.
    This function will also write the "data file" needed by SAMMY

    Parameters
    ----------
    filename_abbrev : str
        The full file path to the file that is being written to, WITHOUT the extension
        (since two files are written with default extensions)
    energy : array-like
        The 1-d vector of energy. This should be the same length as all
        the other pointwise quantities given.
    obs : array-like
        The observable, should be the same length as energy
    obserr : array-like
        The uncertainty on the observable (for convenience). Should be the same
        length as energy, and should be the diagonal of the total covariance matrix.
    stat : array-like
        The statistical covariance matrix, can be 2-d or 1-d as only the diagonal is
        non-zero.
    syst : array-like
        The user can provide the systematic covariance matrix here. It should be 
        noted that **care must be taken to ensure the order of the covariance matches
        the order of derivatives given in sys_der_list**. 
    sys_der_list : array-like
        Each vector in this list is the derivative of the function with 
        respect to that variable. **This is the Jacobian, and the order
        of the derivatives must be the same as the order given in the 
        covariance matrix syst**. 
    high_precision : bool, optional
        If the precision written to the file is too low (default 1e-6), write
        the floats with higher precision (1e-12).

    Returns
    -------
    none : ascii file
        Returns a somewhat annotated file with all the information necessary to
        reproduce the point-by-point covariance of the function, in the format
        SAMMY expects.


    """
    def sixe(x):
        return '{:10.6e}'.format(x)
    if high_precision:
        def sixe(x):
            return '{:10.12e}'.format(x)

    # use numpy 
    energy = np.array(energy)
    stat = np.array(stat)
    syst = np.array(syst)
    sys_der_list = np.array(sys_der_list)
    
    # --- Data file ---
    fmt_twenty([energy,obs,obserr],filename_abbrev+'.twenty')

    # --- IDC file ---
    if len(np.shape(stat))>1:
        stat_err = np.sqrt(np.diag(stat))
    else:
        stat_err = np.sqrt(stat)
    with open(filename_abbrev+'.idc','w+') as f:
        f.write('Number of data-reduction parameters = '+str(len(sys_der_list))+'\n')
        f.write('\n')
        f.write('Free-format partial derivatives\n')
        for i in range(len(sys_der_list[0])):
            f.write(sixe(energy[i])+' '+sixe(stat_err[i])+' ')
            for j in range(len(sys_der_list)):
                f.write(sixe(sys_der_list[j][i]))
                if j != len(sys_der_list)-1:
                    f.write(' ')
            f.write('\n')
        f.write('\n')
        f.write('uncertainties on data-reduction parameters\n')
        for i,err in enumerate(np.sqrt(np.diag(syst))):
            f.write(sixe(err))
            if i != len(syst)-1:
                f.write(' ')
        f.write('\n')
        f.write('\n')
        
        off_diagonal = False
        if ( (np.eye(len(syst))*syst - syst).any() > 0):
            off_diagonal = True
            syst_corr = mt.cov_to_corr(syst)
        if off_diagonal:
            f.write('Correlation for data-reduction parameters\n')
            for i,row in enumerate(syst_corr):
                if i==0:
                    continue
                for j,elem in enumerate(row):
                    if j==i:
                        break
                    f.write(sixe(elem))
                    if j != len(row)-1:
                        f.write(' ')
                f.write('\n')
        f.write('\n')

def print_chann(cpts,factor,base_width,t0,FP,uncertainty="0.8"):
    """
    Print the CHANN input for SAMMY resolution function

    Print the "crunch" boundaries in energy [eV], the 
    grouping factor of base bin width, the bin width [ns],
    and the error for for the bin width.

    Parameters
    ----------
    cpts : list
        A list of "crunch" boundaries (tof locations [us]) used
        to group time-of-flight data.
    factor : list
        The factors to group by on each side of the crunch boundaries.
        The len(factor) should be 1 greater than len(cpts).
    base_width : float
        The base width [ns] of the time bins that have been grouped.
    t0 : float
        The time at which neutrons are emitted from a target [us].
    FP : float
        The length of the path neutrons will be traveling, i.e. the
        'flight path' [m]
    uncertainty : str
        The uncertainty on the bin width [ns]

    """

    if (len(factor) < len(cpts)) or (len(factor) == len(cpts)):
        raise ValueError("You should have more grouping factors than cpts")
    if base_width < 0.1:
        print("Warning: Your base width is < 0.1 ns.",
              "Units must be NANOSECONDS.")
    print("--- Paste into SAMMY res. function ---")
    for i,tof in enumerate(cpts[::-1]):
        print("CHANN 0 ",funky.two(tt.tofe(tof-t0,FP)),
              factor[::-1][i]*base_width,uncertainty)
    print("--------------------------------------")
    return 0

def create_sam_inp(inp_list,filename):
    """
    Create a file to automate the interactive input in SAMMY
    which specifies the various input files and energy regions.

    Parameters
    ----------
    inp_list : list
        A list of strings. Each string should be the full line 
        given to SAMMY as input. Some examples of strings are:
        "inp/samndf.inp" 
        "par/samndf.par"
        "data/trans_ta1.dat 200. 300."
    filename : str
        The name of the file that the strings will be written to.
        Each element of inp_list is written to the file and a 
        newline is added after each string.

    Returns
    -------
    no return : ascii file
    """

    with open(filename,'w+') as f:
        for line in inp_list:
            f.write(line+'\n')

def fmt_twenty(data,filename):
    """
    Function designed to take a data array that contains
    transmission or yield data and create a file that is 
    compatible with the SAMMY code TWENTY format

    Parameters
    ----------
    data : numpy array
        Should be in the format of data[0] = energy 1-d
        array, data[1] = trans/yield 1-d array, 
        data[2] = error on t/y 1-d array

    Returns
    -------
    no return : ascii file
    """
    a = np.array(data).T
    np.savetxt(filename,a,fmt="%20.10f",delimiter="") 

def fmt_ten(data,filename,tot=True):
    """
    Function designed to take a data array that contains
    total or capture cross section data and create a file 
    that is compatible with the FITACS data format.

    Parameters
    ----------
    data : numpy array
        Should be in the format of data[0] = energy 1-d
        array, data[1] = trans/yield 1-d array, 
        data[2] = error on t/y 1-d array
    filename : str
        The full file path or relative
    tot : bool
        Whether it is total cross section to be written, 
        if not, capture cross section will be written

    Returns
    -------
    no return : ascii file
    """
    a = np.array(data).T
    np.savetxt(filename,a,fmt="%10.3e",delimiter="")

    with open(filename,'r+') as f:
        file = f.read()
        f.seek(0,0)
        if tot:
            f.write('TOTAL CROSS SECTION\nABSOLUTE UNCERTAINTY\n'+file)
        else:
            f.write('CAPTURE CROSS SECTION\nABSOLUTE UNCERTAINTY\n'+file)

def fmt_par(data,filename):
    """
    Format parameter values into PAR format readable by SAMMY
    
    Parameters
    ----------
    data : array like
        An array with columns of data: energy, \Gamma_\gamma, 
        \Gamma_n, \Gamma_{x1}, \Gamma_{x2}, vary energy, vary
        \Gamma_\gamma, vary \Gamma_n, vary \Gamma_{x1}, vary
        \Gamma_{x2}, spin group. There should be 11 fields
    filename : str
        The file name to write the formatted data into.
        
    Returns
    -------
    no return : ascii file
    """
    
    a = np.array(data).T
    if len(a[0]) < 11:
        ValueError('Requires 11 fields for each resonance energy.')
    with open(filename,'w+') as f:
        for row in a:
            f.write('{:0<11.5f} {:0<10.5f} {:0<10.5f} '.format(row[0],row[1],row[2]))
            f.write('{:0<10.5f} {:0<10.5f} {:1d} '.format(row[3],row[4],int(row[5])))
            f.write('{:1d} {:1d} {:1d} '.format(int(row[6]),int(row[7]),int(row[8])))
            f.write('{:1d} {:1d}\n'.format(int(row[9]),int(row[10])))



def switch_par_flags(par_file,fname_mod,flag_id,flag_val,erange=None):
    """
    Opens par_file and changes the specified flag_id to the
    flag_val. This allows the user to change 0's to 1's or 
    vice versa to turn on/off fitting for each resonce width.

    Parameters
    ----------
    par_file : str
        The address of a PAR file that is in the format typical for
        the code SAMMY.
    fname_mod : str
        String to be added to the end of the par_file name
    flag_id : str
        Either a single str or list of str's that match ve = vary
        energy, vgg = vary gamma gamma, vc1 = vary channel 1, 
        vc2 = vary channel 2, vc3 = vary channel 3.
    flag_val : int
        Either a 0,1,3 for: don't fit, fit, and PUP respectively.
    erange : 1-d array
        An array of length = 2, where erange[0] = lowest energy
        for flags to be changed, and erange[1] = highest energy
        for flags to be changed.

    Returns
    -------
    no return : ascii file
        Creates a new PAR file with only the specified flags changed.
    """
    ve,vgg,vc1,vc2,vc3 = False,False,False,False,False
    if 've' in flag_id:
        ve = True
    if 'vgg' in flag_id:
        vgg = True
    if 'vc1' in flag_id:
        vc1 = True
    if 'vc2' in flag_id:
        vc2 = True
    if 'vc3' in flag_id:
        vc3 = True
    with open(par_file,"r+") as f:
        initial_par = f.readlines()
    with open(par_file+fname_mod,"w+") as f:
        for line in initial_par:
            line_long_enough = False
            inside_erange = False
            if len(line) > 65:
                line_long_enough = True
            if line_long_enough:
                energy = float(''.join(line[0:10]))
                if erange == None:
                    inside_erange = True
                elif energy > erange[0] and energy < erange[1]:
                    inside_erange = True
                if erange == None or inside_erange:
                    line_list = list(line)
                    if ve == True:
                        line_list[56] = str(flag_val)
                    if vgg == True:
                        line_list[58] = str(flag_val)
                    if vc1 == True:
                        line_list[60] = str(flag_val)
                    if vc2 == True:
                        line_list[62] = str(flag_val)
                    if vc3 == True:
                        line_list[64] = str(flag_val)
                    line_join = ''.join(line_list)
                    f.write(line_join)
                else:
                    f.write(line)
            else:
                f.write(line)


def describe_norsum_header(norsum_file,t0,verbose=False,maxE=10000):
    """
    Read the header of a NORSUM file (from RPIXDR) and return a
    description including the crunch boundaries and factors for 
    SAMMY input file.

    Parameters
    ----------
    norsum_file : ascii file
        An ascii file that has the format of NORSUM files that are
        produced by the program RPIXDR.
    t0 : float
        The time zero [us] for the time of flight data give in 
        norsum_file. This along with the flight path that resides
        in norsum_file calculates the energy crunch boundaries.

    Returns 
    -------
    no return : print statement
        A print statement, that includes crunch boundaries and 
        factors that can be copy and pasted into a SAMMY inp 
        file.

    """
    if maxE == 10000:
        print("----\nDefault value of 10 keV for max energy of ")
        print("crunch boundary is being used\n----")
    header_strings = open(norsum_file).readlines()[0:14]
    if verbose == True:
        for string in header_strings:
            print(string.strip())
    print(header_strings[2].strip())
    # collect data
    fp = float(header_strings[11].strip().split()[0])
    bw = float(header_strings[12].strip().split()[0])
    trig = float(header_strings[13].strip().split()[0])
    ncpts = int(header_strings[5].strip().split()[0])
    cpts,cfactors = np.zeros(4),np.zeros(4)
    tofcpts,ecpts = np.zeros(4),np.zeros(4)
    grp_bw = np.zeros(4)
    for i in range(len(cpts)):
        temp_array = header_strings[6+i].strip().split()
        cpts[i],cfactors[i] = float(temp_array[0]),float(temp_array[1])
        print(cfactors[i])
        
        if i==0:
            grp_bw[i] = bw
            tofcpts[i] = cpts[i]*grp_bw[i]
        else:
            grp_bw[i] = grp_bw[i-1]*2**cfactors[i-1]
            tofcpts[i] = (cpts[i]-cpts[i-1])*grp_bw[i]+tofcpts[i-1]
        ecpts[i] = tt.tofe(tofcpts[i]-t0,fp)
    print('FP :         ',fp)
    print('bin width :  ',bw)
    print('triggers :   ',trig)
    if verbose == True:
        print('comp energies : ',ecpts,'\n','comp factors :  ',cfactors)
    print('--------\nSAMMY input: \n--------\n{:9.8f}{:5d}'.format(bw,ncpts))
    print('{:10.2f}{:10.1f}{:10.2f}{:10.1f}{: >10.2f}{:10.1f}{:10.2f}{:10.1f}'.format(ecpts[2],
          grp_bw[3]/bw,ecpts[1],grp_bw[2]/bw,ecpts[0],grp_bw[1]/bw,maxE,grp_bw[0]/bw))

# input: numPars- how many parameters were varied by SAMMY
#        infile- SAMMY.LPT file that includes printed 
#        covariance matrix.
# output: returns string of new covariance matrix file, and 
#         creates new cov. mat. file.
# take LPT file from sammy, that includes covariance matrix
# and return csv file cov matrix
def LPTtoCOV(numPars,infile):
    """
    Uses SAMMY output LPT file to create a covariance matrix
    file.

    SAMMY.LPT file contains correlations and standard deviations
    for the parameters that were allowed to vary. These are used
    to create a covariance matrix in a more usable text file.

    Parameters
    ----------
    numPars : int
        The number of parameters that were allowed to vary in the 
        SAMMY calcualation.
    infile : str
        The file address of the LPT file to be interrogated.

    Returns
    -------
    newfile : str
        File address of the new csv covariance file

    Notes
    -----
    Function is relatively primitive, it will only work with 
    numPars > 15. 
    """
    
    # !! Requires parameters in LPT file to be >15 !!
    if numPars<15:
        print("You have to edit the source to search for")
        print("strings of \"2\" or \"3\" or whichever numbers")
        print("describe the columns of the correlation matrix")
        print("in the LPT file.")
        raise ValueError("Number of parameters < 15.")

    # initialize correlation matrix of zeros
    cormat = np.zeros((numPars,numPars))
    # initialize standard deviation matrix of zeros
    stdmat = np.zeros((numPars,numPars))
    
    
    flag1,flag2,flag3 = False,False,False
    startval,i = 0,0                                 # used to track column position in cor. matrix
    lines = []
    with open(infile) as lpt:
        lines = [line.split() for line in lpt]       # split every line into str's by space
    for line in lines:
        #print(line)
        if "Number" in line and "varied" in line and "parameters" in line:
            number_of_param = int(line[len(line)-1]) # check number of parameters stated in LPT
            print(' '.join(line))
            if number_of_param != numPars:
                raise ValueError("Re-run program with {} parameters".format(number_of_param))
        if flag1 and "MISSING" in line and "LINES" in line:   # if keywords break the loop
            print("Finished.")
            break
        if flag2 and len(line)==0:                   # if end of block of covar's; start new col's
            startval += 15
            flag3 = True
            if i+1 != numPars:
                print("!!!!\nIs the specified number of par's correct?")
                print("Number of parameters found",i+1)
                print("The matrix may be too large!\n!!!!")
            continue
        if flag3 == True:
            flag3 = False                            # need to skip an extra line, continue again
            continue
        if "CORRELATION" and "MATRIX" and  "OUTPUT" in line:          # starts on the output matrix
            print("found it")
            flag1 = True
        if "1" in line and "14" in line and "15" in line and flag1:   # starts on col. definitions
            print("Starting build...")
            flag2 = True
            continue
        if flag2:
            #print(line)
            ll = len(line)
            i = int(line[0])-1                       # i'th row in cormat is line[0]-1=i
            if i>numPars:
                break                                # stop data collection if numPars is exceeded
            for j in range(3,ll):
                cormat[i][startval+j-3] = float(line[j])/100.0 # set proper elem. of cormat to float
            stdmat[i][i] = float(line[1])                      # set proper elem. of stdmat to float
    
    # correlation matrix, so diagonal values=1
    for k in range(numPars):
        cormat[k][k] = 1.0
        
    covmat = np.dot(np.dot(stdmat,cormat),stdmat)    # [cov] = [std]*[cor]*[std]
    print("Matrix dim.: ",np.shape(covmat))
                
    f = open("covmat_{}.dat".format(infile),"w+")
    for line in covmat:
        printline = [str(val) for val in line]
        f.write(",".join(printline)+"\n")

    f.close()
    print("Written in csv to: covmat_{}.dat".format(infile))

    newfile = "covmat_{}.dat".format(infile)

    return newfile                                   # return string filename 
#------------------------------------------------------------------------------
def mini_cov(infile,sel_pars):
    """
    Convert full covariance matrix file to one only made of 
    parameters you choose.

    Takes covariance matrix file generated by LPTtoCOV() and
    converts large matrix into matrix of covariances specific 
    to the user selection so as to reduce the complexity of 
    the matrix. (300x300 matrix --> 10x10 matrix)

    Parameters
    ----------
    infile : text file
        File of covariances, should be formatted the same as
        the output from LPTtoCOV().
    sel_pars : array
        Vector containing the parameter integer ID numbers
        assigned by sammy and listed in the sammy.lpt file 
        that are specific to the varied parameters the user
        is interested in.
    
    Returns
    -------
    select_cov : numpy array
        2-d matrix containing the covariances of the 
        parameters specified by sel_pars. 
    
    """
    
    # check if input filename contains expected string
    if "covmat" not in infile:
        print("Warning, this is not standard output from LPTtoCOV() function.")
    
    # read them into a lower-triang. array
    cov = np.array(pd.read_csv(infile,header=None))
    print("Initial array size: ",np.shape(cov))

    # convert from lower-tri matrix to symmetric full matrix by adding the transpose
    # and subtracting the diagonal elements
    cov = cov + cov.T - np.diag(cov.diagonal())

    # Test each element on whether cov = transposed cov. If symmetric, all elements should
    # be equal, if not, the boolean array will contain a False value
    if f.symmetric2dTest(cov) == False:
        print("!\n!\n!\n!\n!\n!\n!\n")
        print("Warning, information may have been lost during symmetrization of cov. matrix.")
        print("!\n!\n!\n!\n!\n!\n!\n")
    
    # check to see if user has selected a parameter not in matrix
    if True in [np.shape(cov)[0]<val or val<0 for val in sel_pars]:
        raise ValueError("The cov. mat. doesn't contain some of the requested ID's in sel_pars.")
    
    # make a new array of selected parameters
    select_cov = np.array([[cov[row-1][col-1] for row in sel_pars] for col in sel_pars])
    print("New covariance array of {} parameters.".format(np.shape(select_cov)))
    
    return select_cov
#------------------------------------------------------------------------------


import time
import json
import numpy as np
import pandas as pd

__all__ = ['read_and_add','sum_tof','read_rejected','read_ecals','read_gain_adj']

def read_and_add(filename,hist_df,ecal,gain,llduld,wfcoeff,numadc=4,
                 unweighted=False,verbose=False,mca_df=None,aglgroup=False):
    """
    TODO: incorporate gain adjustment
    
    Read the "list-mode" binary data files and add the events
    to the histograms in the Pandas DataFrame `hist_df`

    Parameters
    ----------
    filename : str
        The full file path to the binary file
    hist_df : DataFrame
        Pandas DataFrame object with attributes (or columns) of `tof`, `adc1`,
        `adc2`, `adc3`, `adc4`. This DataFrame is filled with histograms
        for each channel designated by adc numbers in this function. **This 
        histogram DataFrame must be set up properly** for this function to 
        work. That means the `hist_df.tof` vector must have appropriate bin 
        widths.
    ecal : array-like
        A 2-dimensional array of N rows and 3 columns. Each column is for 
        parameter `A0,A1,A2`. The number of rows is determined by the number of
        ADCs in the data file. Accessing `A1` for ADC1, e.g., would be: 
        `ecal[0][1]`
    gain : array-like
        A 1-d array of gain values, one for every ADC. These are calibrated to match 
        value for beginning run set to source calibration.
    llduld : array-like
        A 2-dimensional array of N rows and 2 columns. Column 0 represents the 
        lower-level discriminator (LLD) and column 1 the upper-level discriminator
        (ULD). The number of rows is determined by the number of ADCs in the file
    wfcoeff : array-like
        A 1-d array for weighting function coefficients defined by the GELINA 
        weighting function. There are 8 parameters, :math:`a_i` from :math:`i=-3,4`
    numadc : int, optional
        The number of ADCs listed **in the binary file. This will change the way**
        **the binary file is read.** Often there will be either 2 or 4 ADCs, 4 is
        the default.
    unweighted : bool, optional
        Whether to add unweighted histograms to the DataFrame 
    verbose : bool, optional
        Whether to increase print output
    mca_df : DataFrame
        Pandas DataFrame object with attributes (or columns) of `ed`, `adc1`,
        `adc2`, `adc3`, `adc4`. This DataFrame is filled with histograms
        for each channel designated by adc numbers in this function. **This 
        histogram DataFrame must be set up properly** for this function to 
        work. `ed` is for "energy deposition"
    agl_group : bool
        Whether to group using bin-edges defined AGL-style (`True`), as opposed to 
        setting TOF compression points for bin edge definition RPI-style (`False`)

    Examples
    --------
    >>> import pandas as pd
    >>> import nuctools as nuc
    >>> user_max_tof = 2.5e3 # [us]
    >>> bin_width = 0.001    # [us]
    >>> numbins = int(user_max_tof/bin_width)+1
    >>> bin_width_arr = np.repeat(bin_width,numbins)
    >>> ecal = [[ 144.85,6.4355,1],
                [ 145.66,6.6452,1],
                [ 179.75,6.3147,1],
                [ 56.069,6.3898,1]] # [keV]
    >>> gain = [1,1,1,1]
    >>> llduld = [[16,1100],
                  [13,1160],
                  [12,1150],
                  [30,1150]] # [keV]
    >>> wfcoeff = [68.7,-513.4,1367.7,-1614.6,993.7,-159.5,16.5,-0.568] # [1/MeV]
    >>> filename = "example.lst"
    >>> data = pd.DataFrame({
            'tof'  : np.cumsum(bin_width_arr),
            'adc1' : np.zeros(numbins),
            'adc2' : np.zeros(numbins),
            'adc3' : np.zeros(numbins),
            'adc4' : np.zeros(numbins)
        })
    >>> nuc.read_and_add(filename,data,ecal,gain,llduld,wfcoeff)

    Notes
    -----

    Formula for the energy calibration is given by:

    .. math:: E_d = A_0 + A_1 \\cdot PA^{A_2}

    where :math:`PA` is pulse area and :math:`E_d` is in units of keV. The 
    weighting function is given by [1]:

    .. math:: W(E_d) = \\sum_{i=-3}^{4} a_i \\cdot E_d^i.

    The :math:`PA` is multiplied by the gain, as calibrated to a calibration source, 
    and logged during runs.

    [1] Borella et. al, The use of C6D6 detectors for neutron induced capture 
    cross-section measurements in the resonance region, Nuclear Instruments and
    Methods in Physics Research A 577 (2007) 626–640
    """

    #------------------------
    # Define binary format
    #------------------------
    # dt = [('tcmsb',np.dtype(">u2")),('tclsb',np.dtype(">u2")),('adc1',np.dtype(">u2")),
    #       ('adc2',np.dtype(">u2")),('adc3',np.dtype(">u2")),('adc4',np.dtype(">u2"))]
    dt = [('tcmsb',np.dtype(">u2")),('tclsb',np.dtype(">u2"))]
    for i in range(1,numadc+1):
        dt.append(('adc{}'.format(i),np.dtype(">u2")))
    #------------------------
    # Read single file
    #------------------------
    d = np.fromfile(filename,dtype=dt)
    
    dat = pd.DataFrame({
        'tc': np.left_shift((d['tcmsb'] & 0x01FF).astype(np.uint32),16) + d['tclsb'],
    })
    for i in range(1,numadc+1):
        dat['adc{}'.format(i)] = d['adc{}'.format(i)].astype(float)
    
    #------------------------
    # Bad TOFs
    #------------------------
    initial_num_events = dat.shape[0]
    dat = dat[dat.tc!=0]
    if(verbose):
        print("Rejected for TOF=0:       {:.3f}%; Number counts lost: {}".format(
            (initial_num_events-len(dat))/initial_num_events*100,initial_num_events-len(dat)))

    datdict = {}
    #------------------------
    # for every adc run these functions
    #------------------------
    for i in range(1,numadc+1):
        if( verbose ):
            if( (dat['adc{}'.format(i)]!=0).any() ):
                print("Found events for ADC{}".format(i))
        #------------------------
        # reject coincidence
        #------------------------
        temp = dat[(dat.loc[:,~dat.columns.isin(['tc'])]>0).sum(axis=1)==1]
        #------------------------
        # split into adc's
        #------------------------
        temp = temp[['tc','adc{}'.format(i)]][(temp['adc{}'.format(i)]!=0)].reset_index(drop=True).copy()
        #------------------------
        # save mca before lld and uld cuts (but bin-structure can still cut!) and apply gain
        #------------------------
        if gain[i-1] == 1.0:
            datdict["mca{}".format(i)] = temp['adc{}'.format(i)]
        else:
            # add uniform rand [0,1) if gain != 1 (prevent non-physical zeros)
            rnd = np.random.rand(len(temp['adc{}'.format(i)]))
            temp['adc{}'.format(i)] = gain[i-1]*(temp['adc{}'.format(i)] + rnd)-1
            datdict["mca{}".format(i)] = temp['adc{}'.format(i)]
        #------------------------
        # apply lower and upper level discriminators
        #------------------------
        temp = temp[(temp['adc{}'.format(i)] > llduld[i-1][0]) & (temp['adc{}'.format(i)] < llduld[i-1][1])]
        #------------------------
        # Calibrate pulse area (PA) to energy deposited [keV]
        #------------------------
        temp['Ed'] = (ecal[i-1][0] + ecal[i-1][1]*(temp['adc{}'.format(i)])**ecal[i-1][2])
        #------------------------
        # apply weighting function to get the weighted counts (coeff are [1/MeV])
        #------------------------
        temp['Cw'] = np.zeros(len(temp['Ed']))
        for j in range(-3,4+1):
            temp.Cw += wfcoeff[j+3]*(temp['Ed']/1000)**(j)
        datdict["d{}".format(i)] = temp
        
    #------------------------
    # add to histograms
    #------------------------
    max_tof_ns = hist_df.tof.max() # should be in [ns]
    numbins = len(hist_df)
    bin_def = numbins
    binrange = (0,max_tof_ns)
    if( aglgroup ): 
        last_tof = 2*hist_df.tof[numbins-1]-hist_df.tof[numbins-2]
        bin_def = np.insert(np.array(hist_df.tof),numbins,last_tof)
        bin_range = (0,last_tof)

    #------------------------
    # add weighted
    #------------------------
    for i in range(1,numadc+1):
        hist_df['adc{}'.format(i)] += np.histogram(datdict["d{}".format(i)].tc,weights=datdict["d{}".format(i)].Cw,bins=bin_def)[0]

    #------------------------
    # add unweighted if requested
    #------------------------
    if( unweighted ):
        for i in range(1,numadc+1):
            hist_df['uwadc{}'.format(i)] += np.histogram(datdict["d{}".format(i)].tc,bins=bin_def)[0]

    #------------------------
    # add mca if requested
    #------------------------
    if( mca_df is not None ):
        numbins = len(mca_df)
        for i in range(1,numadc+1):
            mca_df['adc{}'.format(i)] += np.histogram(datdict["mca{}".format(i)],bins=numbins,range=(0,numbins))[0]

    
def sum_tof(agl_inp_file,file_list):
    """
    Sum the TOF histograms from all events in the file list into a Pandas
    DataFrame and return the DataFrame

    Parameters
    ----------
    agl_inp_file: str
        The file name that describes necessary input for reading list-mode 
        files like AGL
    file_list : list
        A list of strings with full filepaths. Suggest to glob with Python: 
        `glob.glob("my/directory/*.lst")`
    
    Returns
    -------
    data : DataFrame
        A Pandas DataFrame with column names: "tof", "adc1", "adc2",
        "adc3", "adc4"
    data,mca : tuple
        A tuple of Pandas DataFrames with columns as above for `data` and columns
        of "ed", "adc1", "adc2", "adc3", "adc4" for `mca`

    Examples
    --------
    >>> import glob
    >>> import nuctools as nuc
    >>> filelist   = glob.glob('La/la_fp4a_t03_r03_*.lst')
    >>> 
    >>> data = nuc.sum_tof("test-files/test-agl-inp.json",filelist)

    """

    # -----------------------------------------------------
    # Read dictionary into memory and move to variables
    # -----------------------------------------------------
    with open(agl_inp_file,'r') as f:
        agldict = json.load(f)
    numadc = agldict['numadc']
    llduld = agldict['llduld']
    mcabins = agldict['mcabins']
    unweighted = agldict['unweighted']
    verbose = agldict['verbose']
    aglgroup = agldict['useAGLgrouping']
    ecal = agldict['ecal']
    gain_adj = agldict['gain_adj']
    wfcoeff = agldict['wfcoeff']
    bin_width = agldict['bin_width']
    user_max_tof = agldict['max_tof']
    cfct = agldict['cfct']
    zones = agldict['zones']
    rundict = agldict['rundict']


    print("Total number of files: ",len(file_list))
    if(verbose):
        print("Number of ADCs: ",numadc)
        print("Number bins for MCA: ",mcabins)

    # ------------------------------
    # create the TOF histo
    # - internally we use only [ns] to allow integer math
    # - TODO: move all input to YAML file
    # ------------------------------
    bin_width *= 1000 # go to [ns]
    if( aglgroup ):
        numbins = np.sum(zones*1024)
        toflist = np.array([0])
        for i,mult in enumerate(cfct):
            zone_array = np.cumsum(np.repeat(2**mult*bin_width,zones[i+1]*1024))
            if i==0:
                toflist = np.insert(zone_array,0,0.0)
            else:
                toflist = np.concatenate([toflist,zone_array+toflist[len(toflist)-1]])
        data = pd.DataFrame({
            'tof'  : toflist,
            'adc1' : np.zeros(len(toflist)),
            'adc2' : np.zeros(len(toflist)),
        })
        data.drop(data.tail(1).index,inplace=True)
    else:
        numbins = int(user_max_tof/bin_width)+1
        bin_width_arr = np.repeat(bin_width,numbins)
        bin_width_arr[0] = 0 # begin first tof bin at zero
        data = pd.DataFrame({
            'tof'  : np.cumsum(bin_width_arr),
        })
    for i in range(1,numadc+1):
        data['adc{}'.format(i)] = np.zeros(numbins)
        if( unweighted ):
            data['uwadc{}'.format(i)] = np.zeros(numbins)
    if( verbose ):
        print("Number bins for TOF: ",numbins)
        print(data.head())
    # ------------------------------
    # create the MCA histo
    # ------------------------------
    mca = pd.DataFrame({
        'ed' : np.linspace(0,np.max(llduld),mcabins)
        })
    for i in range(1,numadc+1):
        mca['adc{}'.format(i)] = np.zeros(mcabins)
    if( mcabins==0 ):
        mca = None
    # ------------------------------
    # Sum over the files in a folder
    # ------------------------------
    rundict_keys = np.sort(list(rundict.keys()))
    start = time.time()
    i,j,k=0,0,0
    for filename in np.sort(file_list):
        bad_run_found = False
        for key in rundict_keys:
            if key in filename:
                for bad_run in rundict[key]:
                    if( "{:0>4d}".format(bad_run) in filename and key in filename ):
                        j+=1
                        bad_run_found = True
        if(bad_run_found):
            continue
        # set gain values to proper run
        for k,key in enumerate(rundict_keys):
            if key in filename:
                gain = gain_adj[k]
        if(verbose):
            print("-------------------------\nFile: {}\n-------------------------".format(filename))
            print("Gain: ",gain)
        if(i==10):
            time_per_file = (time.time()-start)/10
            print("Time/file at 10 files: {:.2f} secs".format(time_per_file))
        if(i==len(file_list)//2):
            print("50% done.")

        read_and_add(filename,data,ecal,gain,llduld,wfcoeff,numadc,unweighted,verbose,mca,aglgroup)
        i+=1
    end = time.time()
    print("    Cycles thrown out: {}/{}".format(j,len(file_list)))
    print("  Total time required: {:.2f} secs".format(end-start))

    if( mcabins>0 ):
        return data,mca
    else:
        return data

def read_rejected(agl_outfile,expid):
    """
    Read existing AGL output file for rejected run IDs 
    and the rejected run, place into string for pyagl

    Parameters
    ----------
    agl_outfile : str
        Full path to AGL report file, typically ending with "*rep.txt"
    expid : str
        unique string for run ids, for example "zr91_fp14a" when the 
        run id strings are "zr91_fp14a_c01_r01", etc.

    Returns
    -------
    json_string : str
        A string in json format that can be copy and pasted into the
        pyagl input file to match the same behavior as AGL

    Examples
    --------
    >>> import nuctools
    >>> agl_file = "ornl_fp14_WF_Fe50mm_2022_BCoNaS_an_rep.txt"
    >>> agl_file_path = "/some/dir/"
    >>> expid = "fe50_fp14a"
    >>> rejected_string = nuc.read_rejected(agl_file_path+agl_file,expid)

    """
    with open(agl_outfile,'r') as f:
        lines = f.readlines()
    rejected_dict = {}
    for line in lines:
        if "Runtime" in line:
            break
        if expid in line:
            runid = line.split()[0]
            rejected_dict[runid] = []
        if "rejected" in line:
            reject_list = line.split()
            reject_list = np.array(reject_list[1:len(reject_list)]).astype(int).tolist()
            rejected_dict[runid] = reject_list
    return json.dumps(rejected_dict,sort_keys=True)

def read_ecals(agl_outfile,numadc,numruns):
    """
    Read the energy calibrations for every adc and every run

    Parameters
    ----------
    agl_outfile : str
        Full path to AGL report file, typically ending with "*rep.txt"
    numadc : int
        The number of ADCs that were used in the experiment
    numruns : int
        The number of runs that have energy calibrations listed. Keep 
        in mind AGL produces 1 extra listed set of parameters 
    """
    with open(agl_outfile,'r') as f:
        lines = f.readlines()
    adcsfound = 0
    runsfound = 0
    numparam = 3
    ecal = np.zeros((numruns,numadc,numparam)) # no error catch for def too small
    # load the ecals for every adc for every run, skip last listed (repeat)
    for line in lines:
        # break if extra run found
        if "1EN_CAL_A0" in line:
            runsfound+=1
            if runsfound > numruns:
                break
        if "EN_CAL_A0" in line:
            adc = int(line[0])
            if adc>adcsfound:
                adcsfound = adc
            a0 = float(line.split()[1])
            ecal[runsfound-1,adc-1,0] = a0
        if "EN_CAL_A1" in line:
            adc = int(line[0])
            if adc>adcsfound:
                adcsfound = adc
            a1 = float(line.split()[1])
            ecal[runsfound-1,adc-1,1] = a1
        if "EN_CAL_A2" in line:
            adc = int(line[0])
            if adc>adcsfound:
                adcsfound = adc
            a2 = float(line.split()[1])
            ecal[runsfound-1,adc-1,2] = a2
    # test if ADCs is unexpected
    if adcsfound != numadc:
        print("ADCs found   = ",adcsfound)
        print("ADCs defined = ",numadc)
        raise ValueError("Number of ADCs defined is not equal to ADCs found.")
    print(ecal)

def read_gain_adj(agl_outfile,numadc,numruns):
    """
    Read the pulse area (MCA) gain adjustments for every adc and every run

    Parameters
    ----------
    agl_outfile : str
        Full path to AGL report file, typically ending with "*rep.txt"
    numadc : int
        The number of ADCs that were used in the experiment
    numruns : int
        The number of runs that have energy calibrations listed. Keep 
        in mind AGL produces 1 extra listed set of parameters 
    """
    with open(agl_outfile,'r') as f:
        lines = f.readlines()
    gain_adj = np.zeros((numruns,numadc))
    runsfound = 0
    for line in lines:
        if "1AMP_GAIN_ADJUSTMENT" in line:
            runsfound += 1
        if runsfound > numruns:
            break
        if "AMP_GAIN_ADJUSTMENT" in line:
            adc = int(line[0])-1
            gain = line.split()[1]
            gain_adj[runsfound-1,adc] = gain
    np.set_printoptions(precision=5)
    print(gain_adj)




import numpy as np
import glob
import time
import warnings
import os



__all__ = ['describe_norsum_header','Rpy_xdr']

def describe_norsum_header(norsum_file,t0):
    x = 1


class Rpy_xdr:
    
    """
    Read all the data files in a specified folder with a specified extension 
    to sum counts over all cycles and run monitor statistics.

    Python program designed to mimic the GUI based analysis program at RPI called
    RPI XDR.
    
    TODO: consider removing read_files() from initialization and making methods
    to read and write files to avoid summing every time.

    TODO: write monitor file

    TODO: include dead time correction with dtc from sample_create

    Parameters
    ----------
    folder : str
        The folder where the raw data files live
    ext : str
        The file extension type (e.g. '.mp')
    mon_line_num : int
        The line number where the monitors are specified. *This is very 
        TOF clock dependent*, but for the files generated at the RPI 
        linac ('*.mp' and '*.889') monitors are listed 5 monitor counts
        per line for 2 lines. The mon_line_num parameter should be set 
        to that first line.
    num_samp : int
        How many samples are contained in the files in the specified 
        folder
    samp_line_num : int
        The line number on which the sample number is specified. This 
        program assumes that the sample number is the third element 
        of an array split by the space character (e.g. 'cmline3=Sample #  1')
    raw_bw : float
        The width of the counting bins in time [us]
    min_tof : float
        The minimum cutoff on the TOF spectrum [us] used calculate the total 
        detector counts in the monitor ratios. Used to cut out the gamma 
        flash


    Attributes
    ----------
    monitors : list
        A list of numpy arrays. Each element is a numpy array which corresponds
        to the 10 monitor counts for each cycle for a given sample.
    det_sum : list
        A list of numpy arrays. Each element is a sum (over cycles) of TOF 
        dependent detector counts for a given sample.
    mon_std : list
        A list of numpy arrays. Each element corresponds to different samples,
        each containing the fractional standard deviation of the monitor to 
        detector sum ratio for all 10 monitors.

    Methods
    -------
    read_files
        Read the files for the specified extension and folder. This function
        reads the files and stores the monitors and det_sum for each sample 
        and cycle.
    mon_frac_std
        Calculates the fractional standard deviation of the monitor to
        detector sum ratios for each sample and stores this information
        in mon_std
    write_tof_files
        Writes the contents of each element (and therefore each sample) of 
        det_sum array to a separate file along with the associated time
        of flight.

    Notes
    -----
    Fractional standard deviations are computed by:

    .. math:: \\frac{\\sigma\\left(\\frac{C_{Mon}}{C_{Det}}\\right)}{\\langle\\frac{C_{Mon}}{C_{Det}}\\rangle}

    where :math:`\\sigma()` is the standard deviation as computed by the numpy package.

    """
    
    def __init__(self,folder,ext='mp',mon_line_num=43,num_samp=None,samp_line_num=40,
                 raw_bw=None,head_len=None,min_tof=0.0,verbose=False,mon_header=None):
        
        self.folder = folder
        self.ext = ext
        self.mon_line_num = mon_line_num
        self.samp_line_num = samp_line_num
        self.head_len = head_len
        self.mon_header = mon_header

        if num_samp is None:
            self.num_samp = 4
            warnings.warn("Setting number of samples to 4")
        else:
            self.num_samp = num_samp
            if verbose:
                print("Number of samples: {}".format(self.num_samp))
        
        if raw_bw is None:
            self.raw_bw = 0.0064
            warnings.warn("Setting bin width to default: 0.0064 us")
        else:
            self.raw_bw = raw_bw

        if head_len is None:
            self.head_len = 59
            warnings.warn("Setting header length to 59")

        if mon_header is None:
            self.mon_header = "m1 m2 m3 m4 m5 m6 m7 m8 m9 m10 det"
            warnings.warn("Setting monitor header to default.")

        self.min_tof = min_tof

        # ---------------------------
        # Failures
        # ---------------------------
        if os.path.isdir(folder) is False:
            raise ValueError("Specified folder does is not a directory.")
        
        # ---------------------------
        # read the files with specified extension
        # ---------------------------
        self.monitors = [None]*self.num_samp  # <-- Each index could be diff. len()
        self.det_sum  = [None]*self.num_samp  # <-- Each index could be diff. len()
        for samp_num in range(self.num_samp):
            self.read_files(samp_num,verbose)
        
        # ---------------------------
        # calculate the percent std of monitor/det_counts ratio
        # ---------------------------
        self.mon_std = [None]*self.num_samp
        for samp_num in range(self.num_samp):
            self.mon_frac_std(samp_num)
    
    def read_files(self,samp_num,verbose):
        
        """
        Grab all the files with the specified extension and
        sample number. Read the monitor counts and detector
        TOF counts and store them into memory
        
        """
        
        first_file = True
        
        for i,file in enumerate(np.sort(glob.glob('{}*.{}'.format(self.folder,self.ext)))):

            if verbose: 
                print("File {}, {}".format(i,file))
            
            # ---------------------------
            # continue if wrong sample
            # ---------------------------
            sample = np.genfromtxt(file,max_rows=1,skip_header=self.samp_line_num-1,comments='*')[2]
            if sample != samp_num:
                continue
            
            # ---------------------------
            # grab the 10 monitors listed on the two rows, flatten to one list
            # ---------------------------
            temp = np.genfromtxt(file,max_rows=2,skip_header=self.mon_line_num-1)[:,2:].flatten()
            
            # ---------------------------
            # read the detector tof counts
            # ---------------------------
            det = np.genfromtxt(file,skip_header=self.head_len)
            
            # ---------------------------
            # gate the tof of total detector counts for monitor normalization
            # ---------------------------
            tof_array = np.arange(len(det))*self.raw_bw
            det_counts = np.sum(det[tof_array>=self.min_tof])
            temp = np.hstack((temp,det_counts))
            
            # ---------------------------
            # place into memory
            # ---------------------------
            if first_file:
                # set the monitors list
                self.monitors[samp_num] = temp
                first_file = False
                # set the detector sum
                self.det_sum[samp_num] = det
            else:
                self.monitors[samp_num] = np.vstack((self.monitors[samp_num],temp))
                self.det_sum[samp_num] += det
        
        # ---------------------------
        # transpose to make first index of that samples monitors the number of cycles
        # ---------------------------
        self.monitors[samp_num] = self.monitors[samp_num].T
        
    def mon_frac_std(self,samp_num):
        """
        Calculate the percent stand. deviation for all monitors of a 
        specific sample: samp_num
        
        """
        # ---------------------------
        # set length to len of monitor array
        # ---------------------------
        self.mon_std[samp_num] = np.empty(len(self.monitors[samp_num]))
        self.mon_std[samp_num][:] = np.nan
        
        # ---------------------------
        # standard deviation of mon/det ratio over mean(mon/det ratio)
        # ---------------------------
        for i,mon in enumerate(self.monitors[samp_num]):
            self.mon_std[samp_num][i] = (np.std(mon/self.monitors[samp_num][10]) / 
                               np.mean(mon/self.monitors[samp_num][10]))
            
    def write_tof_files(self,output_folder,name_string):
        
        """
        Takes the whatever is in the detector sum attribute: det_sum and
        writes it to file.
        
        Parameters
        ----------
        output_folder : str
            The folder to write the data files to
        name_string : str
            The string that should be placed at the beginning of each samples
            data file (experiment title, isotope of interest, etc....)
        
        """
        
        # ---------------------------
        # function overwrites files
        # ---------------------------
        warnings.warn("This function will overwrite old files.")
        
        for samp_num in range(self.num_samp):
            
            # ---------------------------
            # Add tof vector and transpose array to go in columns
            # ---------------------------
            tof_array = np.arange(len(self.det_sum[samp_num]))*self.raw_bw
            save_arr = np.vstack((tof_array,self.det_sum[samp_num])).T

            # ---------------------------
            # Get the monitors
            # ---------------------------
            monstring = ""
            for mon in self.monitors[samp_num].sum(axis=1):
                monstring += "{} ".format(mon)

            # ---------------------------
            # write the monitors to file
            # ---------------------------
            np.savetxt('data/{}_mon_{}.dat'.format(name_string,samp_num),
                       self.monitors[samp_num].T,fmt="%d",
                       header=self.mon_header,comments='')
            
            # ---------------------------
            # Save with 6 pt. precision
            # ---------------------------
            np.savetxt(output_folder+"{}_samp_{}.dat".format(name_string,samp_num),
                       save_arr,fmt='%.6f',header=monstring)
                
                


import numpy as np
import pandas as pd

# -----------------------------------------------------------------------------
# Define some constants that can be used anywhere
# -----------------------------------------------------------------------------
hbar = 6.582119514e-16 # [eV-s] NIST Planck's over 2 pi
mnc2 = 939565413.3  # NIST neutron mass [eV]
c = 299.792458      # NIST speed of light [m/us]
Na = 6.02214076e23  # NIST Avogadro's num. [1/mol]
# -----------------------------------------------------------------------------

__all__ = ['pole_strength',
           'n_strength_func_frohner','n_strength_func','spin_stat_factor',
           'r_inf','r_prime','calc_ac','penetrability','rho','wave_number',
           'wigner_spacing']

def pole_strength(S,A,a_c):
    """
    Pole strength as defined by Frohner
    
    Parameters
    ----------
    S : float
        Neutron strength function (can be calculated with 
        physics_tools). Typical value is 1e-4. [unitless]
    A : float
        The mass number of the nucleus.
    a_c : float,optional
        The nuclear scattering radius of the isotope [fm].
    
    Returns
    -------
    s : float
        The pole strength as defined by Frohner

    Notes
    -----
    .. math:: s_c = \\frac{S_l}{2k_ca_c}\\sqrt{\\frac{E}{1 eV}} 

    .. math:: s_c = \\frac{S_l}{2\\left(\\frac{\\sqrt{2m_nE}}{\\hslash/2\\pi}\\frac{A}{A+1}\\right)a_c}\\sqrt{\\frac{E}{1 eV}} 

    .. math:: s_c = \\frac{S_l(h/(2\\pi))(A+1)}{a_cA\\sqrt{8m_n}}
    """
    
    s = S*hbar*(A+1)/a_c/1e-15/A/np.sqrt(8*mnc2/(c*1e6)**2)
    return s

def n_strength_func_frohner(e_vec,gn_vec,J,L,I,A=None,a_c=None,e_high=None):
    """
    This should really be the same as Mughabghab. This does not need 
    to be used.

    """

    # Is A defined?
    if A is not None:
        if a_c is None:
            a_c = calc_ac(A)
    else:
        raise ValueError('You have not specified A')
    # Vectors match?
    if len(e_vec) != len(gn_vec):
        raise ValueError('e_vec and gn_vec must be same length')

    if e_high is not None:
        vals = pd.DataFrame({
            'e' : e_vec,
            'gn': gn_vec,
            })
        e_vec = np.array(vals.e[vals.e < e_high])
        gn_vec = np.array(vals.gn[vals.e < e_high])
    else:
        e_vec = np.array(e_vec)
        gn_vec = np.array(gn_vec)

    # channel penetrability
    P_c = penetrability(L,a_c,A,e_vec)

    # average reduced neutron width amplitude squared, 1/1000 for meV->eV
    avg_red_width = np.mean(gn_vec/1000/2/P_c)

    # average level spacing
    D = np.mean(np.diff(e_vec))

    # strength function
    g = spin_stat_factor(J,I)
    k = wave_number(A,e_vec)
    S_l = np.mean(2*g*k*a_c/np.sqrt(e_vec)*avg_red_width)/D/(2*L+1)

    return S_l






def n_strength_func(e_vec,gn_vec,J,L,I,A=None,a_c=None,e_high=None):
    """
    Calculate the neutron strength function

    Calculate the neutron strength function as defined
    in Mughabghab's Atlas with the reduced neutron 
    amplitudes.

    Parameters
    ----------
    e_vec : array-like
        A 1-d vector of energy eigenvalues associated with
        each of the resonance amplitudes [eV]
    gn_vec : array-like
        A 1-d vector of the neutron widths [meV]
    J : float or array-like
        J is the compound nucleus spin, and takes on values
        within :math:`\\mid I-L+1/2\\mid \\leq J\\leq\\mid I+L+1/2\\mid`
        where L is the quantum angular momentum.
    L : float
        The quantum angular momentum. Can have only integer values.
    I : float
        I is the spin of the target nuclei and can be integer or half
        integer
    A : float
        The mass number of the nucleus. If a_c is not specified
        the mass number must be specified to calculate a_c
    a_c : float,optional
        The nuclear scattering radius of the isotope [fm]. a_c must
        be specified if A is not specified.
    e_high : float
        The upper energy in which to sum of the resonance values. Sum
        will go from lowest to max(e_vec) or to e_high

    Returns
    -------
    S_l : float
        The neutron strength function [dimensionless]

    Examples
    --------
    >>> import nuctools as nuc
    >>> E  = np.array([5,10,15,20])
    >>> gn = np.array([3,12, 8,21])
    >>> J,L,I,A = 4.0,0,3.5,181
    >>> print(nuc.n_strength_func(E,gn,J,L,I,A=A))

    Notes
    -----
    .. math:: S_l = \\frac{1}{(2l+1)\\Delta E}\\sum_j g_j\\Gamma^l_{n,j}

    where,

    .. math:: \\Gamma^l_{n,j} = \\sqrt{\\frac{1 eV}{E_0}}\\frac{\\Gamma_{n,j}}{V_l}

    where,

    .. math:: V_l = \\frac{P_l}{k_ca_c}

    :math:`k_c` is the wave number and :math:`a_c` is the channel radius.

    Mughabghab, S.F., Atlas of Neutron Resonances, 2006, 5th
    Edition

    """
    # Is A defined?
    if (A is not None) and (a_c is None):
        a_c = calc_ac(A)
    # Is a_c defined?
    elif( a_c is None ):
        raise ValueError('You have not specified a_c or A')
    if (A is None):
        raise ValueError('You have not specified A')
    # Vectors match?
    if len(e_vec) != len(gn_vec):
        raise ValueError('e_vec and gn_vec must be same length')

    if e_high is not None:
        vals = pd.DataFrame({
            'e' : e_vec,
            'gn': gn_vec,
            'J' : J,
            })
        e_vec = np.array(vals.e[vals.e < e_high])
        gn_vec = np.array(vals.gn[vals.e < e_high])
        J = np.array(vals.J[vals.e < e_high])
    else:
        e_vec = np.array(e_vec)
        gn_vec = np.array(gn_vec)
        J = np.array(J)

    #Mughabghab's V_l value
    V_l = penetrability(L,a_c,A,e_vec)/rho(a_c,A,e_vec)
    # Reduced neutron width amplitudes (with extra factors), 1/1000 for meV->eV
    reduced_gn = (gn_vec/1000/e_vec**(1/2.)*
        spin_stat_factor(J,I)/V_l)
    
    S_l = 1/((2*L+1)*(max(e_vec)-min(e_vec)))*reduced_gn.sum()

    N = len(e_vec)

    dS_l = (2/N)**0.5*S_l

    return S_l,dS_l

def spin_stat_factor(J,I):
    """
    The spin statistical factor 
    
    The spin statistical factor is the probability that
    a particular compound state will form for a given J
    and I when neutrons interact with the nucleus.

    Parameters
    ----------
    J : float or array-like
        J is the compound nucleus spin, and takes on values
        within :math:`\\mid I-L+1/2\\mid \\leq J\\leq\\mid I+L+1/2\\mid`
        where L is the quantum angular momentum.
    I : float
        I is the spin of the target nuclei and can be integer or half
        integer

    Returns
    -------
    g : float
        The probability of forming a state with spin compound nucleus
        spin J

    Notes
    -----
    .. math:: g = \\frac{2J+1}{2(2I+1)}
    
    """
    
    g = (2*J+1)/(2*(2*I+1))
    return g

def r_inf(Rprime,a_c=0.0,A=None):
    """
    Calculate the distant level contribution to 
    nuclear scattering radius.

    Calculate :math:`R^{\\infty}` from the neutron channel
    radius (or mass number A) and the effective nuclear 
    radius. This is for s-waves, using the R-matrix 
    formalism. **It is recommended that the user input
    parameter A, and let the program calculate :math:`a_c`.**

    Parameters
    ----------
    Rprime : float
        The effective potential scattering radius [fm] 
        including the distant levels contributions from 
        :math:`R^{\\infty}`.
    a_c : float,optional
        The nuclear scattering radius of the isotope [fm]
    A : float, optional
        The mass number of the nucleus. If a_c is not specified
        the mass number must be specified to calculate a_c

    Returns
    -------
    Rinf : float
        The distant levels contribution to the scattering
        radius [fm]

    Notes
    -----
    .. math:: R' = a_c(1-R^\\infty)

    .. math:: R^\\infty = 1-\\frac{R'}{a_c}

    Mughabghab, S.F., Atlas of Neutron Resonances, 2006, 5th
    Edition

    Examples
    --------
    >>> import nuctools as nuc
    >>> A = 181
    >>> Rprime = 7.8
    >>> Rinf = nuc.r_inf(Rprime,A=181)
    
    """

    if a_c == 0.0:
        if A is None:
            raise ValueError("If a_c is to be default, specify mass number A")
        a_c = calc_ac(A)

    Rinf = 1-Rprime/a_c

    return Rinf

def r_prime(Rinf,a_c=0.0,A=None):
    """
    Calculate the effective nuclear radius.

    Calculate the effective nuclear radius from
    the input scattering radius and the contributions
    from distant levels: :math:`R^{\\infty}`. This is 
    for s-waves, using the R-matrix formalism. **It is
    recommended that the user input parameter A, and 
    let the program calculate :math:`a_c`.**

    Parameters
    ----------
    Rinf : float
        The :math:`R^{\\infty}` term which is the contributions
        of many distant levels to the potential scattering [fm].
    a_c : float,optional
        The nuclear scattering radius of the isotope [fm]
    A : float, optional
        The mass number of the nucleus. If a_c is not specified
        the mass number must be specified to calculate a_c

    Returns
    -------
    R_prime : float
        The effective nuclear scattering radius [fm]

    Notes
    -----
    .. math:: R' = a_c(1-R^\\infty)

    Mughabghab, S.F., Atlas of Neutron Resonances, 2006, 5th
    Edition

    Examples
    --------
    >>> import nuctools as nuc
    >>> A = 181
    >>> Rinf = 0.1
    >>> R_prime = nuc.r_prime(Rinf,A=A)
    
    """

    if a_c == 0.0:
        if A is None:
            raise ValueError("If a_c is to be default, specify mass number A")
        a_c = calc_ac(A)

    R_prime = a_c*(1-Rinf)

    return R_prime

def calc_ac(A):
    """
    Calculate good guess of neutron channel radius.

    Parameters
    ----------
    A : int
        The mass number of the nucleus e.g. Ta has 
        one isotope of mass number 181 (essentially)

    Returns
    -------
    a_c : float
        The channel radius [fm] or [Fermi]

    Notes
    -----
    .. math:: a_c = 1.23A^{1/3} + 0.8

    F.H. Frohner, JEFF Report 18, p. 52

    """

    a_c = 1.23*A**(1/3.) + 0.8
    return a_c

def penetrability(L,a_c,A,E):
    """
    Hard sphere penetrability for neutron.

    Parameters
    ----------
    L : int
        The quantum angular momentum
    a_c : float
        The channel radius, which is the value seperating
        the internal and external wave functions in R-matrix,
        or the nuclear radius [fm] or [Fermi].
    A : float
        The mass number of the target isotope, e.g. neutron
        incident on Ta-181 has A=181.
    E : float
        The kinetic energy of the neutron [eV]

    Returns
    -------
    P_c : float
        The hard sphere penetrability factor

    Examples
    --------
    >>> import nuctools as nuc
    >>> L = 0
    >>> a_c = 7.8  # fm
    >>> A   = 181  # mass number
    >>> E   = 100  # eV
    >>> P_c = nuc.penetrability(L,a_c,A,E)

    """

    # calculate rho
    p = rho(a_c,A,E)


    if (L==0):
        P_c = p
    if L==1:
        P_c = p**3/(1+p**2)
    if L==2:
        P_c = p**5/(9+3*p**2+p**4)
    if L==3:
        P_c = p**7/(225+45*p**2+6*p**4+p**6)
    if L==4:
        P_c = p**9/(11025+1575*p**2+135*p**4+10*p**6+p**8)

    return P_c


def rho(a_c,A,E):
    """
    The dimensionless quantity rho

    Parameters
    ----------
    a_c : float
        The channel radius, which is the value seperating
        the internal and external wave functions in R-matrix,
        or the nuclear radius [fm] or [Fermi].
    A : float
        The mass number of the target isotope, e.g. neutron
        incident on Ta-181 has A=181.
    E : float
        The kinetic energy of the neutron [eV]

    Returns
    -------
    p : float
        The dimensionless quantity rho

    Examples
    --------
    >>> import nuctools as nuc
    >>> p = nuc.rho(7.8,181,100)
    >>> print(p)

    Notes
    -----
    .. math:: \\rho = k_ca_c

    where :math:`k_c` is the wave number and :math:`a_c` is 
    the channel radius
    """

    p = wave_number(A,E)*a_c

    return p

def wave_number(A,E):
    """
    Calculate the neutron wave number.

    Calculate the neutron wave number with the reduced
    mass of the neutron and it's target. Useful for 
    equations involving neutron interactions.

    Parameters
    ----------
    A : float
        The mass number of the target isotope, e.g. neutron
        incident on Ta-181 has A=181.
    E : float
        The kinetic energy of the neutron [eV]

    Returns
    -------
    k : float
        The neutron reduced mass wave number [1/Fermi]

    Examples
    --------

    >>> import nuctools as nuc
    >>> k = nuc.wave_number(181,100)
    >>> print(k)

    Notes
    -----
    .. math:: k_c = \\frac{\\sqrt{2m_nE}}{\\hslash/(2\\pi)}\\frac{A}{A+1}

    where :math:`h` is Planck's constant and :math:`m_n` is neutron
    mass, and A is the mass number.
    
    """

    k = np.sqrt(2*mnc2*E)/(c*1e6)/hbar * A/(A+1) * 1e-15 # 1e-15 [m/Fermi]
    return k

def wigner_spacing(x,avD):
    """
    Calculate the Wigner level-spacing distribution from the average 
    spacing D (`avD`)

    Parameters
    ----------
    x : array-like
        The abscissae for the distribution
    avD : float
        The average level spacing

    Returns
    -------
    wig : array-like
        The y-values of the probability distribution

    Examples
    --------
    >>> import nuctools as nuc
    >>> import numpy as np
    >>> x = np.linspace(0,100,100)
    >>> Px = nuc.wigner_spacing(x,4.0)


    Notes
    -----
    .. math:: P(S) = (\\pi S/2D^2)e^{-\\pi/4 S^2/D^2} dS
    """

    wig = np.pi/2/avD**2*x*np.exp(-1/4*np.pi*x**2/avD**2)
    return wig








from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\\n" + fh.read()


setup(
    name='nuctools',
    version='v1.4.4', # updated in CI/CD
    description="nuclear data reduction tools",
    author="Jesse Brown",
    author_email="brownjm1968@gmail.com",
    packages=['nuctools'],
    include_package_data=True,
    license='BSD 3-clause',
    url='https://github.com/brownjm1968/nuctools',
    long_description_content_type="text/markdown",
    long_description=long_description,
    install_requires=['numpy','pandas','scipy','h5py','pyyaml',
                      'matplotlib','pytest'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ]
)




import nuctools.sam_tools as st
import numpy as np
from numpy.testing import assert_array_equal,assert_allclose
import os


def test_read_pds():
    pds_file = os.getcwd() + '/tests/test_files/sammy.pds'

    df = st.read_pds(pds_file)

    gold = np.array([0.794262,4.302552E-02,10.3511,-0.200283,3.46130,-3.926995E-02,-6.50670, 38.5814,5.76615]) 

    assert_allclose(gold,df.iloc[0])


import numpy as np
import pandas as pd
import nuctools.tof_tools as tt

mnc2 = 939565413.3  # NIST neutron mass [eV]
c = 299.792458      # NIST speed of light [m/us]

def test_tofe():
	tof = 100.0
	FP = 100.0

	E = mnc2*(1/np.sqrt(1-(FP/tof/c)**2)-1)
	assert tt.tofe(tof,FP) == E

def test_etof():
	E = 100.0
	FP = 100.0

	tof = FP/c*1/np.sqrt(1-1/(E/mnc2+1)**2)
	assert tt.etof(E,FP) == tof

def test_single_group():
	# x-axis
	x = np.arange(8)
	# random y
	answer = np.arange(len(x))*np.random.random()
	
	# group size
	gs = 2
	# data to be grouped
	xdata = np.repeat(x,gs)
	data = np.repeat(answer,gs)
	
	# grouped data
	gdata = tt.single_group([xdata,data],gs)

	
	np.testing.assert_allclose(gdata[2]/gs,answer)


import numpy as np



__all__ = ['stat_cov','sys_cov','build_sys_cx',
           'symmetric2dTest','covar','cov_to_corr','prop_err_mean',
           'flat_triang_to_full']


def stat_cov(stat_err,stat_der):
    """
    Statistical covariance from the given errors and derviatives,
    which will always be diagonal as they are uncorrelated. This is
    meant to be used along with sys_cov() to get the full covariance
    matrix.

    Parameters
    ----------
    stat_err : array-like
        The statistical error in the variables. This should be 2-d,
        where each position stat_err[j] is a vector, usually TOF 
        or energy.
    stat_der : array-like
        The derivative of the variable in whichever equation you 
        are building the covariance for. This is 2-d, where each
        position stat_der[j] is a vector with length equal to the
        length of the function you are building covariance for.
        **The order of stat_der should be same as stat_err**.

    Returns
    -------
    cy : numpy array
        The covariance given the input error and derivatives. This
        is an NxN array where N = len(derivative vector)

    Notes
    -----
    This will return a diagonal matrix with each element ii as:

    .. math:: C_{y,ii} = \\sum_j \\left(\\frac{\\partial f}{\\partial j}\\Delta j\\right)^2

    where the derivatives of f are taken with respect to each of the consitituents
    j of the f equation. i goes from 1..N where N = len(f).

    Examples
    --------
    >>> a = np.array([1,2,3,4])
    >>> b = np.array([4,3,2,1])
    >>> da = np.array([0.1,0.2,0.1,0.2])
    >>> db = np.array([0.1,0.2,0.1,0.2])
    >>> f = a/b
    >>> dfda = 1/b
    >>> dfdb = a/b**2
    >>> cy = nuc.stat_cov([da,db],[dfda,dfdb])


    """
    # cast as numpy arrays
    stat_err = np.array(stat_err)
    stat_der = np.array(stat_der)
    # square the error for the variance
    stat_err2 = np.power(stat_err,2)
    # square the derivatives
    stat_der2 = np.power(stat_der,2)
    # sum of (derivative * error)^2
    cy = np.diag((stat_err2*stat_der2).sum(axis=0))
    
    # return the diagonal matrix
    return cy


def sys_cov(sys_err,sys_der,cov_mat=None):
    """
    Systematic covariance from the given errors and derviatives.

    Parameters
    ----------
    sys_err : array-like
        The error associated with each of the systematic variables. We 
        assume that these parameters are not already correlated. The 
        sys_err array is expected to be 1-d, with len(sys_err) = the 
        number of systematic variables are in the equation. As these
        variable do not change over the function, they are not vectors.
    sys_der : array-like
        The derivatives of the function of interest with repsect to 
        each of the variables. Each sys_der[i] should be the length
        of the function of interest. **The order of sys_der should be 
        same as sys_err, with any cov_mat variables plugged into the
        leading rows **.
    cov_mat : array-like
        This is a covariance matrix that can be plugged into the 
        leading rows and columns. This allows already correlated data
        to be used in the systematic error. (sys_err still has to 
        contain variables for now.)

    Returns
    -------
    cy : numpy array
        The covariance given the input error and derivatives.

    Notes
    -----
    The output covariance will be NxN where N = len(function) of interest.
    The algebra to get the covariance is shown below.

    .. math:: \\mathbf{C_y} = \\mathbf{F_x}\\mathbf{C_x}\\mathbf{F_x}^T

    Here the dimensions of F_x are NxM where M is the number of systematic 
    variables (and the len(sys_der)), and N is the len(function) (and the 
    len(sys_der[0])). The matrix C_x is MxM.

    Examples
    --------
    >>> a = np.array([1,2,3,4])
    >>> b = np.array([4,3,2,1])
    >>> c = 3.0
    >>> da = np.array([0.1,0.2,0.1,0.2])
    >>> db = np.array([0.1,0.2,0.1,0.2])
    >>> dc = 0.1
    >>> f = a*(b*c)**-1
    >>> dfdc = a*(b*c)**-2*(-1)*(b)
    >>> cy = sys_cov([dc],[dfdc])

    """

    if cov_mat is not None:
        lcm = len(cov_mat)
        N = lcm + len(sys_err)
        cx = np.zeros((N,N))
        # plug in matrix at the leading rows and cols
        cx[0:lcm,0:lcm] = cov_mat
        
    # cast as numpy arrays
    sys_err = np.array(sys_err)
    sys_der = np.array(sys_der)
    # square the error for the variance and diagonalize
    sys_err2 = np.diag(np.power(sys_err,2))
    # determine the final input covariance matrix
    if cov_mat is None:
        cx = sys_err2
    else:
        cx[lcm:,lcm:] = sys_err2
    # sandwich rule, matrix algebra to get outgoing covariance
    # (@ symbol is matrix dot product)
    cy = sys_der.T@cx@sys_der
    
    # return NxN matrix (N = len(derivatives))
    return cy

def build_sys_cx(sys_err,cov_mat=None):
    """
    Systematic covariance input matrix from the given errors.

    Parameters
    ----------
    sys_err : array-like
        The error associated with each of the systematic variables. We 
        assume that these parameters are not already correlated. The 
        sys_err array is expected to be 1-d, with len(sys_err) = the 
        number of systematic variables are in the equation. As these
        variable do not change over the function, they are not vectors.
    cov_mat : array-like
        This is a covariance matrix that can be plugged into the 
        leading rows and columns. This allows already correlated data
        to be used in the systematic error. (sys_err still has to 
        contain variables for now.)

    Returns
    -------
    cy : numpy array
        The covariance given the input error and derivatives.

    Notes
    -----
    Same as sys_cov(), but only builds the input matrix. Initial purpose
    is for printing the input matrix to a file.

    """
    if cov_mat is not None:
        lcm = len(cov_mat)
        N = lcm + len(sys_err)
        cx = np.zeros((N,N))
        # plug in matrix at the leading rows and cols
        cx[0:lcm,0:lcm] = cov_mat
        
    # cast as numpy arrays
    sys_err = np.array(sys_err)
    # square the error for the variance and diagonalize
    sys_err2 = np.diag(np.power(sys_err,2))
    # determine the final input covariance matrix
    if cov_mat is None:
        cx = sys_err2
    else:
        cx[lcm:,lcm:] = sys_err2

    return cx

def symmetric2dTest(matrix2d):
    """
    Test 2-dimensional matrix on whether it is symmetric

    Parameters
    ----------
    matrix2d: numpy array
        2-dimensional array to be tested for symmetry

    Returns
    -------
    symmBool: Boolean 
        True or False value whether matrix is symmetric
    """
    
    # is the matrix 2-d?
    if len(np.shape(matrix2d)) != 2:
        raise ValueError("Matrix dimensions are not equal to 2.")
    matrix2d = np.array(matrix2d)

    # create boolean for whether 2-d matrix = its transpose
    symmBool = (matrix2d == matrix2d.T).all()
            

    if symmBool == False:
        print("Matrix not symmetric.")
        print("Max assymetry = ",np.max(matrix2d-matrix2d.T))

    return symmBool




def covar(fx,cx):
    """
    Find covariance of a function whose inputs covariance matrix
    are in cx, and the Jacobian of the function with respect to
    it's input variables is in fx.

    Parameters
    ----------
    fx : numpy array
        The sensitivity matrix (or Jacobian) that has
        N rows (corresponding to the number of function
        values) and M columns (the number of input values
        to the function). NxM matrix.
    cx : numpy array
        The input covariance matrix which corresponds to 
        fx. This should include every input (a_1,b_1,a_2,
        b_2...) which is differentiated in fx. MxM matrix.


    Returns
    -------
    cy : numpy array
        The resultant covarianc matrix due to the input 
        covariance matrix cx, and sensitivities of the 
        given function in fx. NxN matrix.

    Examples
    --------
    >>> import nuctools as nuc
    >>> fx = np.array([[1,2],[1,2],[1,2]])
    >>> cx = np.array([[2,0],[0,2]])
    >>> print(fx.shape,"x",cx.shape,"x",fx.T.shape)
    >>> cy = nuc.covar(fx,cx)
    >>> print(cy.shape)

    Notes
    -----
    For independent variables at x_i:

    .. math:: a_i, b_i

    .. math:: C_y = F_xC_xF_x^T

    .. math:: f(x_i) = \\frac{a_i}{b_i}

    The covariance between two points in the function are:

    .. math:: covar(f(x_1),f(x_2)) = F_xC_xF_x^T

    .. math:: F_x = \\left[ \\begin{array}{cccc} \\frac{\\partial f(x_1)}{\\partial a_1} & \\frac{\\partial f(x_1)}{\\partial b_1} & \\frac{\\partial f(x_1)}{\\partial a_2} & \\frac{\\partial f(x_1)}{\\partial b_2} \\cr \\frac{\\partial f(x_2)}{\\partial a_1} & \\frac{\\partial f(x_2)}{\\partial b_1} & \\frac{\\partial f(x_2)}{\\partial a_2} & \\frac{\\partial f(x_2)}{\\partial b_2} \\end{array} \\right]
    
    .. math:: C_x = \\left[ \\begin{array}{cccc} \\Delta a_1^2 & 0 & 0 & 0 \\cr 0 & \\Delta b_1^2 & 0 & 0 \\cr 0 & 0 & \\Delta a_2^2 & 0 \\cr 0 & 0 & 0 & \\Delta b_2^2  \\end{array} \\right]

    """
    
    fx = np.array(fx)
    cx = np.array(cx)
    
    shape_fx = fx.shape
    shape_cx = cx.shape
    
    
    if shape_fx[1] != shape_cx[0]:
        print('-----------------------------------------')
        print("Shapes of fx and cx cannot be multiplied:")
        print(shape_fx,"x",shape_cx)
        print('-----------------------------------------')
        raise ValueError('Input matrices are not compliant')
    
    cy = np.dot(np.dot(fx,cx),fx.T)
    
    print("Size of Cy matrix: ",np.shape(cy))
    
    return cy

def cov_to_corr(cy):
    """
    Convert covariance matrix to it's correlation matrix

    Parameters
    ----------
    cy : numpy array
        The input covariance matrix. Should be square and
        symmetric.

    Returns
    -------
    corr : numpy array
        The corresponding correlation matrix

    Examples
    --------
    >>> import nuctools as nuc
    >>> fx = np.array([[1,2],[1,2],[1,2]])
    >>> cx = np.array([[2,0],[0,2]])
    >>> print(fx.shape,"x",cx.shape,"x",fx.T.shape)
    >>> cy = nuc.covar(fx,cx)
    >>> print(cy.shape)
    >>> corr = nuc.cov_to_corr(cy)

    Notes
    -----
    The relationship of correlation to covariance is:

    .. math:: \\frac{cov(X,Y)}{sd(X)sd(Y)}

    """
    
    N = len(cy)
    
    corr = np.zeros((N,N))
    
    sd = np.sqrt(np.diag(cy))
    
    sdinv = np.diag(1/sd)
    
    corr = np.dot(np.dot(sdinv,cy),sdinv)
    
    #print(np.shape(corr))
    return corr


def prop_err_mean(dx,axis=None):
    """
    Calculate the propagated error for the mean of values
    whose error is known. Only observed to work for 1-d 
    2-d arrays.


    Parameters
    ----------
    dx : 1-d array
        The error on the values which are input for the mean

    Returns
    -------
    dmean : numpy array
        The propagated error on the mean given the error on 
        the averaged values.

    Examples
    --------
    >>> import nuctools as nuc
    >>> vals = np.array([1,2,3])
    >>> vals_error = np.array([0.1,0.1,0.1])
    >>> mean = np.mean(vals)
    >>> error_of_mean = nuc.prop_err_mean(vals_error)

    Notes
    -----
    If axis is specified, the average can be used for 2-d 
    arrays in axis-0,1.

    .. math:: \\Delta\\langle x\\rangle = \\frac{1}{N}\\sqrt{\\sum_i \\Delta x_i^2}

    """
    dx = np.array(dx)
    if axis is not None:
        if axis == 0:
            N = len(dx)
        if axis == 1:
            N = len(dx[0])
    else:
        N = len(dx)
    dx2 = np.power(dx,2)
    dmean = 1/N*np.sqrt(dx2.sum(axis=axis))
    return dmean

def flat_triang_to_full(triang):
    """
    Take a flat 1-dim array that represents upper-triang
    matrix and return the full symmetric matrix

    Parameters
    ----------
    triang : array-like
        Should be 1-dimensional, upper-triangular row-major order

    Returns
    -------
    full : array-like
        The full symmetric array

    Examples
    --------
    >>> import nuctools as nuc
    >>> import numpy as np
    >>> lt = np.array([1,3,5,4,9,2])
    >>> nuc.flat_triang_to_full(lt)
    array([[1., 3., 5.],
          [3., 4., 9.],
          [5., 9., 2.]])

    Notes
    -----
    Future changes may be needed to allow options for upper/lower 
    and row-/column-major ordering of the flat array

    """
    rows = int(1/2*(np.sqrt(8*len(triang)+1)-1))
    mask = np.tri(rows,dtype=bool).T # transpose: lower to upper
    full = np.zeros((rows,rows))
    full[mask] = triang
    full += full.T*np.tri(rows,k=-1)
    return full







from nuctools import *
import numpy as np
from numpy.testing import assert_array_equal,assert_allclose
import os
import filecmp


def test_read_endf_float():

	gold_float = 7e4

	assert read_endf_float(" 7.000000+4") == gold_float

def test_write_endf_float():

	gold_string = "-7.000000+4"

	assert write_endf_float(-7e4) == gold_string

def test_read_pendf_xs():
	answer = np.array([[1.0,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8],
		               [1.5,1.5,1.5,1.5,1.5,1.5,1.5,1.5,1.5]])

	file = os.getcwd() + '/tests/test_files/test_pendf.dat'
	start = 98
	finish = 100
	ptwise_cross_sec = read_pendf_xs(file,start,finish)

	assert_array_equal(answer,ptwise_cross_sec)


def test_write_pendf_xs():
	energy = [1e2,2e2,3e3,4e3,5e3,6e3,7e3,8e3,9e3,1e4]
	cs = [2,2,2,2,2,2,2,2,2,2]
	material_number = 7328
	file_number = 3
	reaction_number = 1
	test_file = os.getcwd() + '/tests/test_files/temp_pendf_xs.dat'
	ans_file = os.getcwd() + '/tests/test_files/test_pendf_write.dat'

	write_pendf_xs(test_file,energy,cs,material_number,file_number,reaction_number)

	assert filecmp.cmp(test_file,ans_file) == True




import numpy as np
import pandas as pd
import nuctools.funky as f

mnc2 = 939565413.3  # NIST neutron mass [eV]
c = 299.792458      # NIST speed of light [m/us]




def test_threshFinder():
	pulse = [0,1,2,4,6,8,10,9,8,7,6,5,4,3,2,1,0]
	threshold = 2.0
	answer = 3 #---^

	assert f.threshFinder(pulse,threshold) == answer

def test_decayFinder():
	pulse = [0,1,2,4,6,8,10,9,8,7,6,5,4,3,2,1,0]
	threshold = 2.0
	answer = 15 #---------------------------^

	assert f.decayFinder(pulse,threshold) == answer

def test_giveIt():
	assert f.giveIt() == 900

